# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Operations:
    def abs(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms the absolute value of a variable.

        APDL Command: ABS

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name for identifying the variable on
            the printout and displays.  Embedded blanks are compressed
            upon output.

        facta
            Scaling factor (positive or negative) applied to variable
            IA (defaults to 1.0).

        Notes
        -----
        The new variable is calculated as:

        IR = | FACTA x IA |

        For a complex number (a + ib), the absolute value is the
        magnitude, where the IA values are obtained from:
        ``sqrt(a**2 + b**2)``

        See POST26 - Data Operations in the Mechanical APDL Theory
        Reference for details.
        """
        return self.run(f"ABS,{ir},{ia},,,{name},,,{facta}", **kwargs)

    def add(
        self,
        ir="",
        ia="",
        ib="",
        ic="",
        name="",
        facta="",
        factb="",
        factc="",
        **kwargs,
    ):
        """Adds (sums) variables.

        APDL Command: ADD

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia, ib, ic
            Reference numbers of the three variables to be operated on.  If
            only two variables, leave IC blank.  If only one, leave IB and IC
            blank.

        name
            Thirty-two character name for identifying the variable on the
            printout and displays.  Embedded blanks are compressed upon output.

        facta, factb, factc
            Scaling factors (positive or negative) applied to the corresponding
            variables (default to 1.0).

        Notes
        -----
        Adds variables (up to three at once) according to the operation:

        IR = (FACTA x IA) + (FACTB x IB) + (FACTC x IC)
        """
        command = f"ADD,{ir},{ia},{ib},{ic},{name},,,{facta},{factb},{factc}"
        return self.run(command, **kwargs)

    def atan(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms the arctangent of a complex variable.

        APDL Command: ATAN

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia
            Reference number of the complex variable to be operated on.

        name
            Thirty-two character name for identifying the variable on the
            printout and displays.  Embedded blanks are compressed upon output.

        facta
            Scaling factor (positive or negative) applied to variable
            IA (defaults to 1.0).  Usually FACTA should be set to 1.
            FACTA may affect the position of the angle by a multiple
            of π, resulting in a quadrant change.

        Notes
        -----
        Forms the arctangent of a complex variable according to the
        operation:

        IR = ATAN(FACTA X b/a)

        where a and b are the real and imaginary parts, respectively,
        of the complex variable IA (which is of the form a + ib).  The
        arctangent represents the phase angle (in radians), and is
        valid only for a harmonic analysis (ANTYPE,HARMIC).

        Since the scaling factor is applied uniformly to b/a, applying
        any positive or negative scaling factor will not affect the
        size of the phase angle, with the exception that a negative
        scaling factor will change the results quadrant by : π.  The
        magnitude of a complex number is still obtained through the
        ABS command.  See POST26 - Data Operations in the Mechanical
        APDL Theory Reference for details.
        """
        command = f"ATAN,{ir},{ia},,,{name},,,{facta}"
        return self.run(command, **kwargs)

    def clog(self, ir="", ia="", name="", facta="", factb="", **kwargs):
        """Forms the common log of a variable

        APDL Command: CLOG

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name for identifying the variable on
            printouts and displays.  Embedded blanks are compressed
            for output.

        facta
            Scaling factor applied to variable IA (defaults to 1.0).

        factb
            Scaling factor (positive or negative) applied to the operation
            (defaults to 1.0).

        Notes
        -----
        Forms the common log of a variable according to the operation:

        ``IR = FACTB*LOG(FACTA x IA)``
        """
        return self.run(f"CLOG,{ir},{ia},,,{name},,,{facta},{factb}", **kwargs)

    def conjug(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms the complex conjugate of a variable.

        APDL Command: CONJUG

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name for identifying the variable on printouts
            and displays.  Embedded blanks are compressed for output.

        facta
            Scaling factor (positive or negative) applied to variable (default
            to 1.0).

        Notes
        -----
        Used only with harmonic analyses (ANTYPE,HARMIC).
        """
        return self.run(f"CONJUG,{ir},{ia},,,{name},,,{facta}", **kwargs)

    def deriv(self, ir="", iy="", ix="", name="", facta="", **kwargs):
        """Differentiates a variable.

        APDL Command: DERIV

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        iy, ix
            Reference numbers of variables to be operated on.  IY is
            differentiated with respect to IX.

        name
            Thirty-two character name for identifying the variable on
            printouts and displays. Embedded blanks are compressed for
            output.

        facta
            Scaling factor (positive or negative) applied as shown
            below (defaults to 1.0).

        Notes
        -----
        Differentiates variables according to the operation:

        IR = FACTA x d(IY)/d(IX)
        """
        return self.run(f"DERIV,{ir},{iy},{ix},,{name},,,{facta}", **kwargs)

    def exp(self, ir="", ia="", name="", facta="", factb="", **kwargs):
        """Forms the exponential of a variable.

        APDL Command: EXP

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]). If this number is the same as
            for a previously defined variable, the previously defined
            variable will be overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name for identifying the variable on
            the printout and displays. Embedded blanks are compressed
            upon output.

        facta
            Scaling factor applied to variable IA (defaults to 1.0).

        factb
            Scaling factor (positive or negative) applied to the operation
            (defaults to 1.0).

        Notes
        -----
        Forms the exponential of a variable according to the operation:

        ``IR = FACTB*EXP(FACTA x IA)``
        """
        return self.run(f"EXP,{ir},{ia},,,{name},,,{facta},{factb}", **kwargs)

    def filldata(self, ir="", lstrt="", lstop="", linc="", value="", dval="", **kwargs):
        """Fills a variable by a ramp function.

        APDL Command: FILLDATA

        Parameters
        ----------
        ir
            Define data table as variable IR (2 to NV [NUMVAR]).

        lstrt
            Start at location LSTRT (defaults to 1).

        lstop
            Stop at location LSTOP (defaults to maximum location as determined
            from data previously stored.

        linc
            Fill every LINC location between LSTRT and LSTOP (defaults to 1).

        value
            Value assigned to location LSTRT.

        dval
            Increment value of previous filled location by DVAL and assign sum
            to next location to be filled (may be positive or negative.)

        Notes
        -----
        Locations may be filled continuously or at regular intervals (LINC).
        Previously defined data at a location will be overwritten.
        """
        command = f"FILLDATA,{ir},{lstrt},{lstop},{linc},{value},{dval}"
        return self.run(command, **kwargs)

    def imagin(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms an imaginary variable from a complex variable.

        APDL Command: IMAGIN

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name for identifying the variable on
            the printout and displays.  Embedded blanks are compressed
            upon output.

        facta
            Scaling factor (positive or negative) applied to variable IA
            (defaults to 1.0).

        Notes
        -----
        This command forms a new variable from a complex variable by
        storing the imaginary part as the real part.  The imaginary
        part can then be used in other operations.  Used only with
        harmonic analyses (ANTYPE,HARMIC).

        Complex variables are stored in two-column arrays with the
        real component stored in the first column and the imaginary
        component stored in the second column.  This command extracts
        the value stored in the second column (i.e., imaginary
        component).  However, with harmonic analyses, all variables
        are stored in two-column arrays as complex variables.  If the
        variable is not complex, then the same value is stored in both
        columns.  This command will extract the variable in the second
        column of the array, even if this variable is not the
        imaginary component of a complex variable.
        """
        return self.run(f"IMAGIN,{ir},{ia},,,{name},,,{facta}", **kwargs)

    def int1(
        self,
        ir="",
        iy="",
        ix="",
        name="",
        facta="",
        factb="",
        const="",
        **kwargs,
    ):
        """Integrates a variable.

        APDL Command: INT1

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.
            Table values represent integrated sum of IY to current
            table position of IX.

        iy, ix
            Integrate variable IY with respect to IX.

        name
            Thirty-two character name for identifying the variable on
            the printout and displays.  Embedded blanks are compressed
            upon output.

        facta, factb
            Scaling factors (positive or negative) applied to the
            corresponding variables (default to 1.0).

        const
            Initial value.

        Notes
        -----
        Integrates variables according to the operation:

        IR = ∫ (FACTA x IY) d(FACTB x IX) + CONST
        """
        command = f"INT1,{ir},{iy},{ix},,{name},,,{facta},{factb},{const}"
        return self.run(command, **kwargs)

    def large(
        self,
        ir="",
        ia="",
        ib="",
        ic="",
        name="",
        facta="",
        factb="",
        factc="",
        **kwargs,
    ):
        """Finds the largest (the envelope) of three variables.

        APDL Command: LARGE

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia, ib, ic
            Reference numbers of the three variables to be operated on.  If
            only two, leave IC blank.  If only one, leave IB blank also.

        name
            Thirty-two character name for identifying the variable on the
            printout and displays.  Embedded blanks are compressed upon output.

        facta, factb, factc
            Scaling factors (positive or negative) applied to the corresponding
            variables (default to 1.0).

        Notes
        -----
        Creates a new variable by finding the largest of up to three variables
        according to the operation:

        IR = Largest of (FACTA x IA, FACTB x IB, FACTC x IC)

        The comparison is done at each time location, so that the new variable
        is the "envelope" of the three existing variables.
        """
        command = f"LARGE,{ir},{ia},{ib},{ic},{name},,,{facta},{factb},{factc}"
        return self.run(command, **kwargs)

    def nlog(self, ir="", ia="", name="", facta="", factb="", **kwargs):
        """Forms the natural log of a variable.

        APDL Command: NLOG

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name identifying the variable on
            printouts and displays.  Embedded blanks are compressed
            for output.

        facta
            Scaling factor applied to variable IA (defaults to 1.0).

        factb
            Scaling factor (positive or negative) applied to the operation
            (defaults to 1.0).

        Notes
        -----
        Forms the natural log of a variable according to the operation:

        ``IR = FACTB*LN(FACTA x IA)``
        """
        return self.run(f"NLOG,{ir},{ia},,,{name},,,{facta},{factb}", **kwargs)

    def prod(
        self,
        ir="",
        ia="",
        ib="",
        ic="",
        name="",
        facta="",
        factb="",
        factc="",
        **kwargs,
    ):
        """Multiplies variables.

        APDL Command: PROD

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia, ib, ic
            Reference numbers of the three variables to be operated
            on.  If only two leave IC blank.  If only one, leave IB
            blank also.

        name
            Thirty-two character name identifying the variable on
            printouts and displays.  Embedded blanks are compressed
            for output.

        facta, factb, factc
            Scaling factors (positive or negative) applied to the
            corresponding variables (default to 1.0).

        Notes
        -----
        Multiplies variables (up to three at once) according to the
        operation:

        ``IR = (FACTA x IA) x (FACTB x IB) x (FACTC x IC)``
        """
        return self.run(
            f"PROD,{ir},{ia},{ib},{ic},{name},,,{facta},{factb},{factc}",
            **kwargs,
        )

    def quot(self, ir="", ia="", ib="", name="", facta="", factb="", **kwargs):
        """Divides two variables.

        APDL Command: QUOT

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia, ib
            Reference numbers of the two variables to be operated on.

        name
            Thirty-two character name identifying the variable on printouts and
            displays.  Embedded blanks are compressed for output.

        facta, factb
            Scaling factors (positive or negative) applied to the corresponding
            variables (default to 1.0).

        Notes
        -----
        Divides two variables according to the operation:

        IR = (FACTA x IA)/(FACTB x IB)
        """
        return self.run(f"QUOT,{ir},{ia},{ib},,{name},,,{facta},{factb}", **kwargs)

    def realvar(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms a variable using only the real part of a complex variable.

        APDL Command: REALVAR

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name identifying the variable on printouts and
            displays.  Embedded blanks are compressed for output.

        facta
            Scaling factor (positive or negative) applied to variable IA
            (defaults to 1.0).

        Notes
        -----
        Forms a variable using only the real part of a variable.  Used only
        with harmonic analyses (ANTYPE,HARMIC).

        Complex variables are stored in two-column arrays with the
        real component stored in the first column and the imaginary
        component stored in the second column.  This command extracts
        the value stored in the first column (i.e., real component).
        However with harmonic analyses, all variables are stored in
        two-column arrays as complex variables.  If the variable is
        not complex, then the same value is stored in both columns.
        This command will extract the variable in the first column of
        the array, even if this variable is not the real component of
        a complex variable.
        """
        command = f"REALVAR,{ir},{ia},,,{name},,,{facta}"
        return self.run(command, **kwargs)

    def small(
        self,
        ir="",
        ia="",
        ib="",
        ic="",
        name="",
        facta="",
        factb="",
        factc="",
        **kwargs,
    ):
        """Finds the smallest of three variables.

        APDL Command: SMALL

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting
            variable (2 to NV [NUMVAR]).  If this number is the same
            as for a previously defined variable, the previously
            defined variable will be overwritten with this result.

        ia, ib, ic
            Reference numbers of the three variables to be operated
            on.  If only two, leave IC blank.  If only one, leave IB
            blank also.

        name
            Thirty-two character name identifying the variable on
            printouts and displays.  Embedded blanks are compressed
            for output.

        facta, factb, factc
            Scaling factors (positive or negative) applied to the
            corresponding variables (defaults to 1.0).

        Notes
        -----
        Finds the smallest of three variables according to the operation:

        ``IR = min(FACTA x IA, FACTB x IB, FACTC x IC)``
        """
        command = f"SMALL,{ir},{ia},{ib},{ic},{name},,,{facta},{factb},{factc}"
        return self.run(command, **kwargs)

    def sqrt(self, ir="", ia="", name="", facta="", **kwargs):
        """Forms the square root of a variable.

        APDL Command: SQRT

        Parameters
        ----------
        ir
            Arbitrary reference number assigned to the resulting variable (2 to
            NV [NUMVAR]).  If this number is the same as for a previously
            defined variable, the previously defined variable will be
            overwritten with this result.

        ia
            Reference number of the variable to be operated on.

        name
            Thirty-two character name identifying the variable on printouts and
            displays.  Embedded blanks are compressed for output.

        facta
            Scaling factor (positive or negative) applied to variable IA
            (defaults to 1.0).

        Notes
        -----
        Forms the square root of a variable according to the operation:
        ``IR=sqrt(FACTA*IA)``
        """
        return self.run(f"SQRT,{ir},{ia},,,{name},,,{facta}", **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module to manage downloading and parsing the FEM from the MAPDL gRPC server."""
from functools import wraps
import os
import re
import time
from typing import Dict
import weakref

from ansys.api.mapdl.v0 import ansys_kernel_pb2 as anskernel
import numpy as np

from ansys.mapdl.core.common_grpc import DEFAULT_CHUNKSIZE, parse_chunks
from ansys.mapdl.core.mapdl_grpc import MapdlGrpc
from ansys.mapdl.core.misc import requires_package, supress_logging, threaded

TMP_NODE_CM = "__NODE__"


def requires_model(output=None):
    def decorator(method):
        """
        This function wrap some methods to check if the model contains elements or nodes.
        """

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            if self._has_nodes and self._has_elements:
                return method(self, *args, **kwargs)
            else:
                if not output or output == "array":
                    return np.array([])
                elif output == "list":
                    return []
                elif output == "dict":
                    return {}
                else:  # pragma: no cover
                    raise ValueError("Output type not allowed.")

        return wrapper

    return decorator


class MeshGrpc:
    """Provides an interface to the gRPC mesh from MAPDL."""

    def __init__(self, mapdl):
        """Initialize grpc geometry data"""
        if not isinstance(mapdl, MapdlGrpc):  # pragma: no cover
            raise TypeError("Must be initialized using MapdlGrpc class")
        self._mapdl_weakref = weakref.ref(mapdl)
        mapdl._log.debug("Attached MAPDL object to MapdlMesh.")

        self.logger = mapdl._log
        self._log = mapdl._log

        self._ignore_cache_reset = False
        self._reset_cache()

    def __repr__(self):
        txt = "ANSYS Mesh\n"
        txt += f"  Number of Nodes:              {len(self.nnum)}\n"
        txt += f"  Number of Elements:           {len(self.enum)}\n"
        txt += f"  Number of Element Types:      {len(self.ekey)}\n"
        txt += f"  Number of Node Components:    {len(self.node_components)}\n"
        txt += f"  Number of Element Components: {len(self.element_components)}\n"
        return txt

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    @property
    def _surf(self):
        """External surface"""
        if self._surf_cache is None:
            self._surf_cache = self._grid.extract_surface()
        return self._surf_cache

    @property
    def _has_nodes(self):
        """Returns True when has nodes"""
        # if isinstance(self._nodes, np.ndarray):
        # return bool(self._nodes.size)
        return self.nodes.size != 0

    @property
    def _has_elements(self):
        """Returns True when geometry has elements"""
        return self._elem.size != 0

    def _set_log_level(self, level):
        """Wraps set_log_level"""
        self._mapdl._set_log_level(level)

    def _reset_cache(self):
        """Reset entire mesh cache"""
        if not self._ignore_cache_reset:
            self.logger.debug("Resetting cache")

            self._cache_elem = None
            self._cache_elem_off = None
            self._cache_element_desc = None
            self._cache_nnum = None
            self._cached_elements = None  # cached list of elements
            self._chunk_size = None
            self._elem = None
            self._elem_off = None
            self._enum = None  # cached element numbering
            self._esys = None  # cached element coordinate system
            self._etype = None
            self._etype_cache = None  # cached ansys element type numbering
            self._etype_id = None  # cached element type id
            self._grid = None
            self._grid_cache = None
            self._keyopt = {}
            self._mtype = None  # cached ansys material type
            self._nnum = None
            self._node_angles = None  # cached node angles
            self._node_coord = None  # cached node coordinates
            self._rcon = None  # cached ansys element real constant
            self._rdat = None
            self._rnum = None
            self._secnum = None  # cached section number
            self._surf_cache = None
            self._tshape = None
            self._tshape_key = None

    def _update_cache(self):
        """Threaded local cache update.

        Used when needing all the geometry entries from MAPDL.
        """
        self.logger.debug("Updating cache")
        # elements must have their underlying nodes selected to avoid
        # VTK segfaul
        with self._mapdl.save_selection:
            self._mapdl.nsle("S", mute=True)

            # not thread safe
            self._update_cache_elem()

            threads = [
                self._update_cache_element_desc(),
                self._update_cache_nnum(),
                self._update_node_coord(),
            ]

            for thread in threads:
                thread.join()

            # must occur after read
            self._ignore_cache_reset = True

            # somehow requesting path seems to help windows avoid an
            # outright segfault prior to running CMSEL
            if os.name == "nt":
                _ = self._mapdl.path

            # TODO: flaky
            time.sleep(0.05)

        self._ignore_cache_reset = False

    @threaded
    def _update_cache_nnum(self):
        if self._cache_nnum is None:
            self.logger.debug("Updating nodes cache")
            nnum = self._mapdl.get_array("NODE", item1="NLIST")
            self._cache_nnum = nnum.astype(np.int32)
        if self._cache_nnum.size == 1:
            if self._cache_nnum[0] == 0:
                self._cache_nnum = np.empty(0, np.int32)

    @property
    def _nnum(self):
        """Return node number cache"""
        self._update_cache_nnum().join()
        return self._cache_nnum

    @_nnum.setter
    def _nnum(self, value):
        self._cache_nnum = value

    @threaded
    def _update_cache_element_desc(self):
        if self._cache_element_desc is None:
            self.logger.debug("Updating elements (desc) cache")
            self._cache_element_desc = self._load_element_types()

    @property
    def _ekey(self):
        """Element key description"""
        self._update_cache_element_desc().join()

        # convert to ekey format
        if self._cache_element_desc:
            ekey = []
            for einfo in self._cache_element_desc:
                ekey.append(einfo[:2])
            return np.vstack(ekey).astype(np.int32)
        return np.array([])

    @threaded
    def _update_node_coord(self):
        if self._node_coord is None:
            self._node_coord = self._load_nodes()

    @property
    def _ans_etype(self):
        """FIELD 1 : element type number"""
        if self._etype_cache is None:
            self._etype_cache = self._elem[self._elem_off[:-1] + 1]
        return self._etype_cache

    @property
    def local(self):
        return self._mapdl._local

    @property
    @requires_model()
    def et_id(self):
        """Element type id (ET) for each element."""
        if self._etype_id is None:
            etype_elem_id = self._elem_off[:-1] + 1
            self._etype_id = self._elem[etype_elem_id]
        return self._etype_id

    @property
    @requires_model()
    def tshape(self):
        """Tshape of contact elements."""
        if self._tshape is None:
            shape_elem_id = self._elem_off[:-1] + 7
            self._tshape = self._elem[shape_elem_id]
        return self._tshape

    @property
    @requires_model("dict")
    def tshape_key(self):
        """Dict with the mapping between element type and element shape.

        TShape is only applicable to contact elements.
        """
        if self._tshape_key is None:
            self._tshape_key = np.unique(np.vstack((self.et_id, self.tshape)), axis=1).T
        return {elem_id: tshape for elem_id, tshape in self._tshape_key}

    @property
    @requires_model()
    def material_type(self):
        """Material type index of each element in the archive."""
        # FIELD 0 : material reference number
        if self._mtype is None:
            self._mtype = self._elem[self._elem_off[:-1]]
        return self._mtype

    @property
    @requires_model()
    def etype(self):
        """Element type of each element.

        This is the ansys element type for each element.

        Notes
        -----
        Element types are listed below.  Please see the APDL Element
        Reference for more details:

        https://www.mm.bme.hu/~gyebro/files/vem/ansys_14_element_reference.pdf
        """
        if self._etype is None:
            arr = np.empty(self._ekey[:, 0].max() + 1, np.int32)
            arr[self._ekey[:, 0]] = self._ekey[:, 1]
            self._etype = arr[self._ans_etype]
        return self._etype

    @property
    @requires_model()
    def section(self):
        """Section number"""
        if self._secnum is None:
            self._secnum = self._elem[self._elem_off[:-1] + 3]  # FIELD 3
        return self._secnum

    @property
    @requires_model()
    def element_coord_system(self):
        """Element coordinate system number"""
        if self._esys is None:
            self._esys = self._elem[self._elem_off[:-1] + 4]  # FIELD 4
        return self._esys

    @property
    @requires_model("list")
    def elem(self):
        """List of elements containing raw ansys information.

        Each element contains 10 items plus the nodes belonging to the
        element.  The first 10 items are:

        - FIELD 0 : material reference number
        - FIELD 1 : element type number
        - FIELD 2 : real constant reference number
        - FIELD 3 : section number
        - FIELD 4 : element coordinate system
        - FIELD 5 : death flag (0 - alive, 1 - dead)
        - FIELD 6 : solid model reference
        - FIELD 7 : coded shape key
        - FIELD 8 : element number
        - FIELD 9 : base element number (applicable to reinforcing elements only)
        - FIELDS 10 - 30 : The nodes belonging to the element in ANSYS numbering.

        """
        if self._cached_elements is None:
            self._cached_elements = np.split(self._elem, self._elem_off[1:-1])
        return self._cached_elements

    @property
    def element_components(self):
        """Element components for the archive.

        Output is a dictionary of element components.  Each entry is an
        array of MAPDL element numbers corresponding to the element
        component.  The keys are element component names.
        """
        return self._mapdl.components._get_all_components_type("ELEM")

    @property
    def node_components(self):
        """Node components for the archive.

        Output is a dictionary of node components.  Each entry is an
        array of MAPDL node numbers corresponding to the node
        component.  The keys are node component names.
        """
        return self._mapdl.components._get_all_components_type("NODE")

    @property
    @requires_model()
    def elem_real_constant(self):
        """Real constant reference for each element.

        Use the data within ``rlblock`` and ``rlblock_num`` to get the
        real constant datat for each element.
        """
        # FIELD 2 : real constant reference number
        if self._rcon is None:
            self._rcon = self._elem[self._elem_off[:-1] + 2]
        return self._rcon

    @property
    def ekey(self):
        """Element type key

        Array containing element type numbers in the first column and
        the element types (like SURF154) in the second column.

        """
        return self._ekey

    @property
    def rlblock(self):
        """Real constant data from the RLBLOCK."""
        return self._parse_rlist()

    @property
    def rlblock_num(self):
        """Indices from the real constant data"""
        return list(self._parse_rlist().keys())

    @property
    def nnum(self) -> np.ndarray:
        """Array of currently selected node numbers.

        Examples
        --------
        >>> mapdl.mesh.nnum
        array([    1,     2,     3, ..., 19998, 19999, 20000])
        """
        return self._nnum

    @property
    def nnum_all(self) -> np.ndarray:
        """Array of all node numbers, even those not selected.

        Examples
        --------
        >>> mapdl.mesh.nnum_all
        array([    1,     2,     3, ..., 19998, 19999, 20000])
        """
        self._ignore_cache_reset = True
        with self._mapdl.save_selection:
            self._mapdl.nsel("all", mute=True)

            nnum = self._mapdl.get_array("NODE", item1="NLIST")
            nnum = nnum.astype(np.int32)
            if nnum.size == 1:
                if nnum[0] == 0:
                    nnum = np.empty(0, np.int32)

        self._ignore_cache_reset = False

        return nnum

    @property
    def enum_all(self) -> np.ndarray:
        """Array of all element numbers, even those not selected.

        Examples
        --------
        >>> mapdl.mesh.enum_all
        array([    1,     2,     3, ..., 19998, 19999, 20000])
        """
        self._ignore_cache_reset = True
        with self._mapdl.save_selection:
            self._mapdl.esel("all", mute=True)

            enum = self._mapdl.get_array("ELEM", item1="ELIST")
            enum = enum.astype(np.int32)
            if enum.size == 1:
                if enum[0] == 0:
                    enum = np.empty(0, np.int32)

        self._ignore_cache_reset = False
        return enum

    @property
    @supress_logging
    def n_node(self) -> int:
        """Number of currently selected nodes in MAPDL.

        Examples
        --------
        >>> mapdl.mesh.n_node
        7217
        """
        return int(self._mapdl.get(entity="NODE", item1="COUNT"))

    @property
    @supress_logging
    def n_elem(self) -> int:
        """Number of currently selected elements in MAPDL.

        Examples
        --------
        >>> mapdl.mesh.n_elem
        1520
        """
        return int(self._mapdl.get(entity="ELEM", item1="COUNT"))

    @property
    def node_angles(self):
        """Not yet implemented"""
        return

    @property
    def enum(self) -> np.ndarray:
        """Element numbers of currently selected elements

        Examples
        --------
        >>> mapdl.mesh.enum
        array([    1,     2,     3, ...,  9998,  9999, 10000])
        """
        if self._enum is None:
            if self._mapdl.mesh.n_elem == 0:
                return np.array([], dtype=np.int32)
            else:
                self._enum = self._mapdl.get_array("ELEM", item1="ELIST").astype(
                    np.int32
                )
        return self._enum

    @property
    def key_option(self):
        """Key options of selected element types."""
        self._update_cache_element_desc().join()

        key_opt = {}
        for einfo in self._cache_element_desc:
            keyopt = einfo[2:20]
            if keyopt.any():
                # convert to fortran/ANSYS format
                valid_keyopt = keyopt[keyopt.astype(np.bool_)]
                ans_keyopt = np.vstack((np.nonzero(keyopt)[0] + 1, valid_keyopt))
                key_opt[einfo[0]] = ans_keyopt.T.tolist()

        return key_opt

    @property
    def nodes(self) -> np.ndarray:
        """Array of nodes in Global Cartesian coordinate system.

        Examples
        --------
        >>> mapdl.mesh.nodes
        array([[0.   0.   0.  ]
               [1.   0.   0.  ]
               [0.25 0.   0.  ]
               ...,
               [0.75 0.5  3.5 ]
               [0.75 0.5  4.  ]
               [0.75 0.5  4.5 ]]
        """
        self._update_node_coord().join()
        if self._node_coord is None:
            return np.empty(0)
        return self._node_coord

    @property
    def nodes_in_current_CS(self) -> np.ndarray:
        """Returns the nodes array in the current coordinate system."""
        CS_id = self._mapdl.get_value("active", 0, "CSYS")
        return self.nodes_in_coordinate_system(CS_id=CS_id)

    def nodes_in_coordinate_system(self, CS_id):
        """Return nodes in the desired coordinate system."""
        if CS_id == 0:
            return self.nodes
        else:
            self._mapdl.parameters["__node_loc__"] = self.nodes
            self._mapdl.vfun("__node_loc_cs__", "local", "__node_loc__", CS_id)
            return self._mapdl.parameters["__node_loc_cs__"]

    @property
    def nodes_rotation(self):
        """Returns an array of node rotations"""
        return self._mapdl.nlist(kinternal="").to_array()[:, 4:]

    def _load_nodes(self, chunk_size=DEFAULT_CHUNKSIZE):
        """Loads nodes from server.

        Parameters
        ----------
        chunk_size : int
            Size of the chunks to request from the server.  Default
            256 kB

        Returns
        -------
        np.ndarray
            Numpy array of nodes
        """
        if self._chunk_size:
            chunk_size = self._chunk_size

        request = anskernel.StreamRequest(chunk_size=chunk_size)
        chunks = self._mapdl._stub.Nodes(request)
        nodes = parse_chunks(chunks, np.double).reshape(-1, 3)
        return nodes

    def _update_cache_elem(self):
        """Update the element and element offset cache"""
        if self._cache_elem is None:
            (
                self._cache_elem,
                self._cache_elem_off,
            ) = self._load_elements_offset()

    @property
    def _elem(self):
        """Contingious array of elements.

        Array of elements, with each element starting at the indices
        in offset.  Each element contains 10 items plus the nodes
        belonging to the element.
        """
        self._update_cache_elem()
        return self._cache_elem

    @_elem.setter
    def _elem(self, value):
        self._cache_elem = value

    @property
    def _elem_off(self):
        """Element offset array"""
        self._update_cache_elem()
        return self._cache_elem_off

    @_elem_off.setter
    def _elem_off(self, value):
        self._cache_elem_off = value

    def _load_elements_offset(self, chunk_size=DEFAULT_CHUNKSIZE):
        """Loads elements from server

        Parameters
        ----------
        chunk_size : int, optional
            Size of the chunks to request from the server.

        Returns
        -------
        elements : np.ndarray
            Array of elements, with each element starting at the
            indices in offset.  Each element contains 10 items plus
            the nodes belonging to the element.  The first 10 items
            are:

            mat    - material reference number
            type   - element type number
            real   - real constant reference number
            secnum - section number
            esys   - element coordinate system
            death  - death flag (0 - alive, 1 - dead)
            solidm - solid model reference
            shape  - coded shape key
            elnum  - element number
            baseeid- base element number (applicable to reinforcing
                     elements only)
            nodes  - The nodes belonging to the elementoffset

        offset : np.ndarray
            Array of indices indicating the start of each element.

        """
        if self._chunk_size:
            chunk_size = self._chunk_size

        request = anskernel.StreamRequest(chunk_size=chunk_size)
        chunks = self._mapdl._stub.LoadElements(request)
        elem_raw = parse_chunks(chunks, np.int32)

        if len(elem_raw) == 0:  # for empty mesh.
            return np.array([]), np.array([])
        n_elem = elem_raw[0]

        # ignore zeros
        elem_off_raw = elem_raw[:n_elem]
        elem_off_raw = elem_off_raw[elem_off_raw != 0]
        # TODO: arrays from gRPC interface should include size of the elem array
        lst_value = np.array(elem_raw.size - n_elem, np.int32)
        offset = np.hstack((elem_off_raw - n_elem, lst_value))

        # overwriting the last column to include element numbers
        elems_ = elem_raw.copy()  # elem_raw is only-read
        elems_ = elems_[n_elem:]
        indx_elem = offset[:-1] + 8
        elems_[indx_elem] = self.enum
        return elems_, offset

    def _load_element_types(self, chunk_size=DEFAULT_CHUNKSIZE):
        """Loads element types from the MAPDL server.

        Returns
        -------
        element_types : np.ndarray
            Array of numpy element types.

        Parameters
        ----------
        int
            Size of the chunks to request from the server.
        """
        request = anskernel.StreamRequest(chunk_size=chunk_size)
        chunks = self._mapdl._stub.LoadElementTypeDescription(request)
        data = parse_chunks(chunks, np.int32)
        n_items = data[0]
        split_ind = data[1 : 1 + n_items]
        # empty items sometimes...
        split_ind = split_ind[split_ind != 0]
        return np.split(data, split_ind)[1:]

    @property
    def grid(self):
        """VTK representation of the underlying finite element mesh.

        Examples
        --------
        Store the finite element mesh as a VTK UnstructuredGrid.

        >>> grid = mapdl.mesh.grid
        UnstructuredGrid (0x7f99b4135760)
          N Cells:	32198
          N Points:	50686
          X Bounds:	-1.181e+00, 1.181e+00
          Y Bounds:	-2.362e-01, 0.000e+00
          Z Bounds:	-2.394e+00, 2.509e+00
          N Arrays:	10

        Plot this grid.

        >>> grid.plot()

        Access the node numbers of grid.

        >>> grid.point_data
        Contains keys:
            ansys_node_num
            vtkOriginalPointIds
            origid
            VTKorigID

        >>> grid.point_data['ansys_node_num']
        pyvista_ndarray([    1,     2,     3, ..., 50684, 50685, 50686],
                        dtype=int32)

        Save this grid to disk

        >>> grid.save('grid.vtk')

        Load this grid externally with ParaView or with pyvista

        >>> import pyvista
        >>> pyvista.read('grid.vtk')

        """
        return self._grid

    @property
    @requires_package("pyvista")
    def _grid(self):
        if self._grid_cache is None:
            self._update_cache()
            self._grid_cache = self._parse_vtk(force_linear=True)
        return self._grid_cache

    @_grid.setter
    def _grid(self, value):
        self._grid_cache = value

    def save(
        self,
        filename,
        binary=True,
        force_linear=False,
        allowable_types=None,
        null_unallowed=False,
    ):
        """Save the geometry as a vtk file

        Parameters
        ----------
        filename : str, pathlib.Path
            Filename of output file. Writer type is inferred from
            the extension of the filename.

        binary : bool, optional
            If ``True``, write as binary, else ASCII.

        force_linear : bool, optional
            This parser creates quadratic elements if available.  Set
            this to True to always create linear elements.  Defaults
            to False.

        allowable_types : list, optional
            Allowable element types.  Defaults to all valid element
            types in ``ansys.mapdl.reader.elements.valid_types``

            See ``help(ansys.mapdl.reader.elements)`` for available element types.

        null_unallowed : bool, optional
            Elements types not matching element types will be stored
            as empty (null) elements.  Useful for debug or tracking
            element numbers.  Default False.

        Notes
        -----
        Binary files write much faster than ASCII and have a smaller
        file size.
        """
        grid = self._parse_vtk(
            allowable_types=allowable_types,
            force_linear=force_linear,
            null_unallowed=null_unallowed,
        )
        if grid:
            return grid.save(str(filename), binary=binary)
        else:
            raise ValueError("The mesh is empty, hence no file has been written.")

    def _parse_vtk(
        self,
        allowable_types=None,
        force_linear=False,
        null_unallowed=False,
        fix_midside=True,
        additional_checking=False,
    ):
        from ansys.mapdl.core.mesh.mesh import _parse_vtk

        return _parse_vtk(
            self,
            allowable_types,
            force_linear,
            null_unallowed,
            fix_midside,
            additional_checking,
        )

    def _parse_rlist(self) -> Dict[int, float]:
        # mapdl.rmore(*list)
        with self._mapdl.force_output:
            rlist = self._mapdl.rlist()

        # removing ueless part
        rlist = rlist.replace(
            """   *****MAPDL VERIFICATION RUN ONLY*****
     DO NOT USE RESULTS FOR PRODUCTION
""",
            "",
        )
        constants_ = re.findall(
            r"REAL CONSTANT SET.*?\n\n", rlist + "\n\n", flags=re.DOTALL
        )

        const_ = {}
        for each in constants_:
            values = [0 for i in range(18)]
            set_ = int(re.match(r"REAL CONSTANT SET\s+(\d+)\s+", each).groups()[0])
            limits = (
                int(re.match(r".*ITEMS\s+(\d+)\s+", each).groups()[0]),
                int(re.match(r".*TO\s+(\d+)\s*", each).groups()[0]),
            )
            values_ = [float(i) for i in each.strip().splitlines()[1].split()]

            if not set_ in const_:
                const_[set_] = values

            for i, jlimit in enumerate(range(limits[0] - 1, limits[1])):
                const_[set_][jlimit] = values_[i]

        return const_



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Component related module"""
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    get_args,
)
import warnings
import weakref

import numpy as np
from numpy.typing import NDArray

from ansys.mapdl.core import Mapdl
from ansys.mapdl.core.errors import ComponentDoesNotExits, ComponentIsNotSelected

if TYPE_CHECKING:  # pragma: no cover
    import logging

ENTITIES_TYP = Literal[
    "NODE", "NODES", "ELEM", "ELEMS", "ELEMENTS", "VOLU", "AREA", "LINE", "KP"
]

VALID_ENTITIES = list(get_args(ENTITIES_TYP))

SELECTOR_FUNCTION = [
    "NSEL",
    "NSEL",
    "ESEL",
    "ESEL",
    "ESEL",
    "VSEL",
    "ASEL",
    "LSEL",
    "KSEL",
]

ENTITIES_MAPPING = {
    entity.upper(): func for entity, func in zip(VALID_ENTITIES, SELECTOR_FUNCTION)
}


ITEMS_VALUES = Optional[Union[str, int, List[int], NDArray[Any]]]
UNDERLYING_DICT = Dict[str, ITEMS_VALUES]

warning_entity = (
    "Assuming a {default_entity} selection.\n"
    "It is recommended you use the following notation to avoid this warning:\n"
    ">>> mapdl.components['{key}'] = '{default_entity}' {value}\n"
    "Alternatively, you disable this warning using:\n"
    ">>> mapdl.components.default_entity_warning=False"
)


def _check_valid_pyobj_to_entities(
    items: Union[Tuple[int, ...], List[int], NDArray[Any]]
):
    """Check whether the python objects can be converted to entities.
    At the moment, only list and numpy arrays of ints are allowed.
    """

    if not all([isinstance(each, (int, np.integer)) for each in items]):
        raise ValueError("Only integers are allowed for component definition.")


class Component(tuple):
    """Component object

    Object which contain the definition of a component.

    Parameters
    ----------
    type_ : str
        The entity type. For instance "NODES", "KP", "VOLU", etc

    items_ : None, str, int, List[int], np.array[int]]
        Item ids contained in the component.

    Examples
    --------
    To create a component object

    >>> mycomp = Component("NODES", [1, 2, 3])
    >>>  mycomp
    Component(type='NODES', items=(1, 2, 3))

    To index the content of the component object:

    >>> mycomp[0]
    1
    >>> mycomp[1]
    2

    To access the type of the entities:

    >>> mycomp.type
    "NODES"

    Convert to list:

    >>> list(mycomp)
    [1, 2, 3]
    """

    def __init__(self, *args, **kwargs):
        """Component object"""
        # Using tuple init because using object.__init__
        # For more information visit:
        # https://stackoverflow.com/questions/47627298/how-is-tuple-init-different-from-super-init-in-a-subclass-of-tuple
        tuple.__init__(*args, **kwargs)

    def __new__(
        cls,
        type_: ENTITIES_TYP,
        items_: Tuple[str, Union[Tuple[int], List[int], NDArray[np.int_]]],
    ):
        if not isinstance(type_, str) or type_.upper() not in VALID_ENTITIES:
            raise ValueError(
                f"The value '{type_}' is not allowed for 'type' definition."
            )
        obj = super().__new__(cls, items_)
        obj._type: ENTITIES_TYP = type_

        return obj

    def __str__(self):
        tup_str = super().__str__()
        return f"Component(type='{self._type}', items={tup_str})"

    def __repr__(self) -> str:
        tup_str = super().__repr__()
        return f"Component(type='{self._type}', items={tup_str})"

    @property
    def type(self) -> ENTITIES_TYP:
        """Return the type of the component. For instance "NODES", "KP", etc."""
        return self._type

    @property
    def items(self) -> tuple:
        """Return the ids of the entities in the component."""
        return tuple(self)


class ComponentManager:
    """Collection of MAPDL components.

    Notes
    -----

    **Components need to be selected** using
    :attr:`Mapdl.cmsel() <ansys.mapdl.core.Mapdl.cmsel>` before
    being listed in :attr:`Mapdl.components <ansys.mapdl.core.Mapdl.components>`

    Examples
    --------
    Simply list all the *selected* components:

    >>> mapdl.components
    MAPDL Components
    ----------------
    MYCOMP1                          : NODE
    MYCOMP2                          : ELEM

    Get a component:

    >>> mapdl.components["mycomp1"]
    Component(type='NODE', items=(1, 3, 6))

    Set a component specifying the type and items:

    >>> mapdl.components["mycomp3] = "KP", [1,2,3]

    Set a component without specifying the type, by default it is ``NODE``:

    >>> mapdl.components["mycomp4"] = (1, 2, 3)
    /Users/german.ayuso/pymapdl/src/ansys/mapdl/core/component.py:282: UserWarning: Assuming a NODES selection.
    It is recommended you use the following notation to avoid this warning:
    \>\>\> mapdl.components['mycomp3'] = 'NODES' (1, 2, 3)
    Alternatively, you disable this warning using:
    > mapdl.components.default_entity_warning=False
    warnings.warn(

    You can change the default type by changing
    :attr:`Mapdl.components.default_entity <ansys.mapdl.core.Mapdl.components.default_entity>`

    >>> mapdl.component.default_entity = "KP"
    >>> mapdl.component["mycomp] = [1, 2, 3]
    >>> mapdl.component["mycomp"].type
    'KP'

    You can also create a component from the already selected entities:

    >>> mapdl.lsel("S",1, 2)
    >>> mapdl.components["mylinecomp"] = "LINE"
    >>> mapdl.components["mylinecomp"]
    (1, 2)

    Selecting a component and retrieving it:

    >>> mapdl.cmsel("s", "mycomp3")
    >>> mapdl.components["mycomp3"]
    Component(type='KP', items=(1, 2, 3))
    """

    def __init__(self, mapdl: Mapdl) -> None:
        """Component Manager.

        Component manager of an
        :class:`Mapdl instance <ansys.mapdl.core.Mapdl>` instance.

        Parameters
        ----------
        mapdl : ansys.mapdl.core.Mapdl
            Mapdl instance which this class references to.
        """
        from ansys.mapdl.core.mapdl import MapdlBase

        if not isinstance(mapdl, MapdlBase):
            raise TypeError("Must be implemented from MAPDL class")

        self._mapdl_weakref: weakref.ReferenceType[Mapdl] = weakref.ref(mapdl)
        self.__comp: UNDERLYING_DICT = {}
        self._update_always: bool = True
        self._autoselect_components: bool = (
            False  # if True, PyMAPDL will try to select the CM first if it does not appear in the CMLIST output.
        )

        self._default_entity: ENTITIES_TYP = "NODES"
        self._default_entity_warning: bool = True

    @property
    def default_entity(self):
        """Default entity for component creation."""
        return self._default_entity

    @default_entity.setter
    def default_entity(self, value: ENTITIES_TYP):
        if not isinstance(value, str) or value.upper() not in VALID_ENTITIES:
            raise ValueError(
                f"Only the following entities are allowed:\n{', '.join(VALID_ENTITIES)}"
            )
        self._default_entity = value.upper()

    @property
    def default_entity_warning(self):
        """Enables the warning when creating components other than node components without specifying its type."""
        return self._default_entity_warning

    @default_entity_warning.setter
    def default_entity_warning(self, value: bool):
        self._default_entity_warning = value

    @property
    def _mapdl(self) -> Mapdl:
        """Return the weakly referenced instance of mapdl"""
        return self._mapdl_weakref()

    def _log(self) -> "logging.Logger":
        """MAPDL logger."""
        return self._mapdl.logger

    @property
    def logger(self) -> "logging.Logger":
        """Access to the logger."""
        return self._log()

    @property
    def _comp(self) -> UNDERLYING_DICT:
        """Dictionary with components names and types."""
        if self.__comp is None or self._update_always:
            self.__comp = self._mapdl._parse_cmlist()
        return self.__comp

    @_comp.setter
    def _comp(self, value):
        self.__comp = value

    def __getitem__(self, key: str) -> ITEMS_VALUES:
        forced_to_select = False

        if key.upper() not in self._comp and self._autoselect_components:
            # the component will not appear in "cmlist" if it has not been selected
            # previously using "cmsel". This behaviour is a bit different than with
            # parameters for example. However it is consistent with geometry definition.
            # Then this API might look a bit confusing, because it resemble parameters,
            # but requires to select the CM.
            self._mapdl.cmsel("A", key)
            forced_to_select = True

        try:
            cmtype = self._comp[key.upper()]
        except KeyError:
            # testing if the CM exists or not
            # it seems you can use `*get` with components which are not selected.
            # It will return 0 if it does not exists.
            if self._mapdl.get_value("COMP", key, "type"):
                # the component exits, but it might not be selected
                raise ComponentIsNotSelected(
                    f"The component named '{key}' is not selected. Use 'mapdl.cmsel' to select it."
                )
            else:
                raise ComponentDoesNotExits(
                    f"The component named '{key}' does not exist in the MAPDL instance."
                )

        output = self._mapdl._parse_cmlist_indiv(key, cmtype)
        if forced_to_select:
            # Unselect to keep the state of the things as before calling this method.
            self._mapdl.cmsel("U", key)

        return Component(cmtype, output)

    def __setitem__(self, key: str, value: ITEMS_VALUES) -> None:
        if not isinstance(key, str):
            raise ValueError("Only strings are allowed for component names.")

        if isinstance(value, Component):
            # Allowing to use a component object to be set.
            cmname = key
            cmtype = value.type
            cmitems = value

        elif isinstance(value, tuple):
            if len(value) == 2:
                if not isinstance(value[0], str) or not isinstance(
                    value[1], (list, tuple, np.ndarray)
                ):
                    raise ValueError(
                        "Only strings are allowed for the first argument, and a list or numpy array for the second."
                    )

                if value[0].upper() not in VALID_ENTITIES:
                    raise ValueError(
                        f"The value '{value}' is not allowed for 'type' definition."
                    )

                cmname = key
                cmtype = value[0].upper()
                cmitems = value[1]

            else:
                # Asumng default entity
                if self.default_entity_warning:
                    warnings.warn(
                        warning_entity.format(
                            default_entity=self.default_entity,
                            key=key,
                            value=value,
                        ),
                        UserWarning,
                    )

                cmname = key
                cmtype = self.default_entity
                cmitems = value

        elif isinstance(value, str):
            # create a component with the already selected elements
            if value.upper() not in VALID_ENTITIES:
                raise ValueError(
                    f"The value '{value}' is not allowed for 'type' definition."
                )

            cmname = key
            cmtype = value

            self._mapdl.cm(cmname, cmtype)

            return  # Early exit

        elif isinstance(value, (list, np.ndarray)):
            # Assuming we are defining a CM with nodes
            if self.default_entity_warning:
                warnings.warn(
                    warning_entity.format(
                        default_entity=self.default_entity,
                        key=key,
                        value=value,
                    ),
                    UserWarning,
                )

            cmname = key
            cmtype = self.default_entity
            cmitems = value

        else:
            raise ValueError("Only strings or tuples are allowed for assignment.")

        _check_valid_pyobj_to_entities(cmitems)

        # Using context manager to proper save the selections (including CM)
        with self._mapdl.save_selection:
            # Select the cmitems entities
            func = getattr(self._mapdl, ENTITIES_MAPPING[cmtype].lower())
            func(type_="S", vmin=cmitems)

            # create component
            self._mapdl.cm(cmname, cmtype)

        # adding newly created selection
        self._mapdl.cmsel("A", cmname)

    def __repr__(self) -> str:
        """Return the current components in a pretty format"""
        lines = ["MAPDL Components", "----------------"]
        if self._comp:
            for key, item in self._comp.items():
                lines.append("%-32s : %s" % (key, item))

        return "\n".join(lines)

    def __contains__(self, key: str) -> bool:
        """
        Check if a given component name is present in
        :class:`Mapdl.components <ansys.mapdl.core.Mapdl.components.ComponentManager>`.

        Parameters
        ----------
        name : str
            The name of the component to search for.

        Returns
        -------
        bool
            True if the component name exists *and it is selected*, False otherwise.

        """
        return key.upper() in self._comp.keys()

    def __iter__(self):
        """
        Return an iterator over the component names.

        Returns
        -------
        iterator
            Return an iterator over the component names.

        """
        yield from self._comp.keys()

    @property
    def names(self):
        """
        Return a tuple that contains the components.

        Returns
        -------
        tuple
            Return a tuple that contains the component names.

        """
        return tuple(self._comp.keys())

    @property
    def types(self):
        """
        Return the types of the components.

        Returns
        -------
        tuple
            Return a tuple that contains the types of the components.

        """
        return tuple(self._comp.values())

    def items(self):
        """
        Return a view object that contains the name-type pairs for each component.

        Returns
        -------
        dict_items
            Return a view object that contains the name-type pairs for each component.

        """
        return self._comp.items()

    def select(self, names: Union[str, list[str], tuple[str]], mute=False) -> None:
        """Select Select components given their names

        Select components given their names.

        Parameters
        ----------
        names : Union[str, list[str], tuple[str]]
            Name(s) of the components
        mute : bool, optional
            Whether to mute the `/CMSEL` command output or not, by default False.

        """
        if isinstance(names, str):
            names = [names]

        for i, each_name in enumerate(names):
            if i == 0:
                self._mapdl.cmsel("S", each_name, mute=mute)
            else:
                self._mapdl.cmsel("A", each_name, mute=mute)

    def _get_all_components_type(self, type_: ENTITIES_TYP):
        """Returns a dict with all the components which type matches the entity type"""
        dict_ = {}
        for each_comp in self._comp:
            item = self.__getitem__(each_comp)
            i_type_ = item.type
            i_items = item.items
            if i_type_ == type_:
                dict_[each_comp] = i_items
        return dict_


def _parse_cmlist_indiv(
    cmname: str, cmtype: str, cmlist: Optional[str] = None
) -> List[int]:
    # Capturing blocks
    if "NAME" in cmlist and "SUBCOMPONENTS" in cmlist:
        header = r"NAME\s+TYPE\s+SUBCOMPONENTS"

    elif "LIST COMPONENT" in cmlist:
        header = ""

    cmlist = "\n\n".join(
        re.findall(
            r"(?s)" + header + r"\s+(.*?)\s*(?=\n\s*\n|\Z)",
            cmlist,
            flags=re.DOTALL,
        )
    )

    # Capturing items in each block
    rg = cmname.upper() + r"\s+" + cmtype.upper() + r"\s+(.*?)\s*(?=\n\s*\n|\Z)"
    items = "\n".join(re.findall(rg, cmlist, flags=re.DOTALL))

    # Joining them together and giving them format.
    items = items.replace("\n", "  ").split()
    items = [int(each) for each in items]

    return items


def _parse_cmlist(cmlist: Optional[str] = None) -> Dict[str, Any]:
    include_selection = False
    if re.search(r"NAME\s+TYPE\s+SUBCOMPONENTS", cmlist):
        # header
        #  "NAME                            TYPE      SUBCOMPONENTS"
        blocks = re.findall(
            r"(?s)NAME\s+TYPE\s+SUBCOMPONENTS\s+(.*?)\s*(?=\n\s*\n|\Z)",
            cmlist,
            flags=re.DOTALL,
        )
    elif re.search(r"NAME\s+SELE\s+TYPE\s+SUBCOMPONENTS", cmlist):
        blocks = re.findall(
            r"(?s)NAME\s+SELE\s+TYPE\s+SUBCOMPONENTS\s+(.*?)\s*(?=\n\s*\n|\Z)",
            cmlist,
            flags=re.DOTALL,
        )
        # Using `cmlist,all` which also includes selection
        include_selection = True
    elif "LIST ALL SELECTED COMPONENTS":
        blocks = cmlist.splitlines()[1:]
    else:
        raise ValueError("The format of the CMLIST output is not recognaised.")

    cmlist = "\n".join(blocks)

    def extract(each_line, ind):
        return each_line.split()[ind].strip()

    def extract_line(each_line):
        if include_selection:
            return [extract(each_line, 1), extract(each_line, 2)]
        else:
            return extract(each_line, 1)

    return {
        extract(each_line, 0): extract_line(each_line)
        for each_line in cmlist.splitlines()
        if each_line
    }



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class TracePoints:
    def pltrac(
        self,
        analopt="",
        item="",
        comp="",
        trpnum="",
        name="",
        mxloop="",
        toler="",
        option="",
        escl="",
        mscl="",
        **kwargs,
    ):
        """Displays a particle flow or charged particle trace on an element

        APDL Command: PLTRAC
        display.

        Parameters
        ----------
        analopt
            Analysis option

            FLUID - Particle trace in fluid flow (default)

            ELEC - Particle trace in electric field

            MAGN - Particle trace in magnetic field

            EMAG - Particle trace in presence of both electric and magnetic fields

        item
            Label identifying the item to be contoured.  Valid item labels are
            shown in Table 222: PLTRAC - Valid Item and Component Labels
            below.  Some items also require a component label.  If Item is
            blank, display only the path trajectory.

        comp
            Component of the item (if required).  Valid component labels are
            shown in Table 222: PLTRAC - Valid Item and Component Labels below.

        trpnum
            Trace point number for storing trajectory data for use with PATH
            logic. Defaults to 0 (no trajectory path data is stored for further
            processing with PATH logic).

        name
            Name of prefix of array variable. Defaults to TRAC. NamePOIN stores
            trajectory path points for trace point number TRPNum. If Analopt =
            ELEC, MAGN, or EMAG, two additional array parameters, NameDATA and
            NameLABL, store trajectory path data and labels for the same
            TRPNum.

        mxloop
            Maximum number of loops traced by a particle.  Defaults to 25 for
            Opt = FLUID; otherwise, defaults to 1000.

        toler
            Length tolerance used for particle trajectory geometry calculation.
            Valid only for Analopt = ELEC, MAGN, or EMAG.  If particle trace
            appears to terminate inside an element, adjusting the length
            tolerance may be necessary.  Defaults to 1.0 x 10-8.

        option
            Flow trace option:

            0 - Use the undeformed mesh for computing the flow trace.

            1 - Use the deformed mesh for computing the flow trace.

        escl
            Electric field scale factor. Setting this scale factor affects only
            the tracing, not the field solution results. A negative factor
            corresponds to the opposite vector direction. Valid only for
            Analopt = ELEC or EMAG. Defaults to 1.

        mscl
            Magnetic field scale factor. Setting this scale factor affects only
            the tracing, not the field solution results. A negative factor
            corresponds to the opposite vector direction. Valid only for
            Analopt = MAGN or EMAG. Defaults to 1.

        Notes
        -----
        For a specified item, the variation of the item is displayed along the
        particle trace as a color-contoured ribbon.  The TRPOIN command must be
        used to define a point on the trajectory path.  Multiple traces may be
        displayed simultaneously by defining multiple trace points.  Issue the
        TRPLIS command to list the current tracing points.  Issue the TRPDEL
        command to delete tracing points defined earlier.   Use the PAPUT
        command with the POIN option to retrieve the particle trajectory points
        as path points.

        The model must be 3-D for the ELEC, MAGN, and EMAG analysis options.

        Three array parameters are created at the time of the particle trace:
        TRACPOIN, TRACDATA and TRACLABL. These array parameters can be used to
        put the particle velocity and the elapsed time into path form.   The
        procedure to put the arrays into a path named PATHNAME is as follows:

        Not used if Analopt = FLUID.  If working in the GUI, use the "All
        information" option to retrieve information from all three arrays at
        once.

        If OPTION is set to 1, the deformed mesh is based on the displacement
        degrees of freedom UX, UY, and UZ, which must be available in the load
        step.

        Table: 222:: : PLTRAC - Valid Item and Component Labels

        See the Basic Analysis Guide for more information on particle flow and
        charged particle traces.  See Animation in the Basic Analysis Guide for
        information on particle trace animation.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"PLTRAC,{analopt},{item},{comp},{trpnum},{name},{mxloop},{toler},{option},{escl},{mscl}"
        return self.run(command, **kwargs)

    def trpdel(self, ntrp1="", ntrp2="", trpinc="", **kwargs):
        """Deletes particle flow or charged particle trace points.

        APDL Command: TRPDEL

        Parameters
        ----------
        ntrp1, ntrp2, trpinc
            Delete points from NTRP1 to NTRP2 (defaults to NTRP1) in steps of
            TRPINC (defaults to 1).  If NTRP1 = ALL, NTRP2 and TRPINC are
            ignored and all trace points are deleted.  If NTRP1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).

        Notes
        -----
        Deletes particle flow or charged particle trace points defined with the
        TRPOIN command.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"TRPDEL,{ntrp1},{ntrp2},{trpinc}"
        return self.run(command, **kwargs)

    def trplis(self, ntrp1="", ntrp2="", trpinc="", opt="", **kwargs):
        """Lists the particle flow or charged particle trace points.

        APDL Command: TRPLIS

        Parameters
        ----------
        ntrp1, ntrp2, trpinc
            List points from NTRP1 to NTRP2 (defaults to NTRP1) in steps of
            TRPINC (defaults to 1).  If NTRP1 = ALL, NTRP2 and TRPINC are
            ignored and all trace points are listed.  If NTRP1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).

        opt
            Opt = LOC lists the trace point number location (X, Y, Z). Default.

        Notes
        -----
        Lists the particle flow or charged particle trace points in the active
        display coordinate system [DSYS].  Trace points are defined with the
        TRPOIN command.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"TRPLIS,{ntrp1},{ntrp2},{trpinc},{opt}"
        return self.run(command, **kwargs)

    def trpoin(self, x="", y="", z="", vx="", vy="", vz="", chrg="", mass="", **kwargs):
        """Defines a point through which a particle flow or charged particle trace

        APDL Command: TRPOIN
        will travel.

        Parameters
        ----------
        x, y, z
            Coordinate location of the trace point (in the active coordinate
            system).  If X = P, graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI).

        vx, vy, vz
            Particle velocities in the X, Y and Z directions (in the active
            coordinate system).

        chrg
            Particle charge.

        mass
            Particle mass.

        Notes
        -----
        Defines a point through which a particle flow or charged particle trace
        [PLTRAC] will travel.  Multiple points (50 maximum) may be defined
        which will result in multiple flow traces.  Use TRPLIS to list the
        currently defined trace points and TRPDEL to delete trace points.

        The VX, VY, VZ, CHRG, and MASS arguments only apply to charged
        particles.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"TRPOIN,{x},{y},{z},{vx},{vy},{vz},{chrg},{mass}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Contains the element implement of the MAPDL database service.

This allows lower level the access to the elements in the MAPDL database.

"""
import weakref

from ansys.api.mapdl.v0 import mapdl_db_pb2

from ansys.mapdl.core.errors import MapdlRuntimeError

from . import DBDef, MapdlDb, check_mapdl_db_is_alive


class DbElems:
    """
    Abstract mapdl database element class.

    Examples
    --------
    Create a MAPDL database element instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> elems = mapdl.db.elems
    >>> print(elems)
    MAPDL Database Elements
        Number of elements:          64
        Number of selected elements: 64
        Maximum element number:      64

    Return the element information of element 1.

    >>> elems = mapdl.db.elems
    >>> elem_info = elems.get(1)

    Return the nodes belonging to the element.

    >>> elem_info.nodes
    [2, 27, 37, 8]

    Return the element data.

    >>> elem_info.elmdat
    [1, 1, 1, 1, 0, 0, 14, 0, 0, 0]

    """

    def __init__(self, db):
        """Initialize this class."""
        if not isinstance(db, MapdlDb):  # pragma: no cover
            raise TypeError("``db`` must be a MapdlDb instance")
        self._db_weakref = weakref.ref(db)
        self._itelm = -1

    @property
    def _db(self):
        """Return the weakly referenced instance of the database."""
        return self._db_weakref()

    def __str__(self):
        """Return the string representation of this class."""
        lines = ["MAPDL Database Elements"]
        lines.append(f"    Number of elements:          {self.num()}")
        lines.append(f"    Number of selected elements: {self.num(selected=True)}")
        lines.append(f"    Maximum element number:      {self.max_num}")
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    @check_mapdl_db_is_alive
    def first(self, ielm=0):
        """
        Get the number of the first element.

        This starts at ``inod``, defaults to the first element in the model.

        Parameters
        ----------
        ielm : int, optional
            The first element number to consider as the "first element".

        Returns
        -------
        int
            The first element number after ``ielm``.

        Examples
        --------
        Return the first selected element.

        >>> elems = mapdl.db.elems
        >>> elems.first()
        1

        Return the first element after element 10.

        >>> elems.first(ielm=10)
        11

        """
        self._itelm = ielm
        return self.next()

    @check_mapdl_db_is_alive
    def next(self):
        """
        Return the number of the next selected element.

        You must first call :func:`DbElems.first`.

        Returns
        -------
        int
            The next selected element number. Returns 0 if there are no more nodes.

        Examples
        --------
        Call :func:`DbElems.first` first.

        >>> elems = mapdl.db.elems
        >>> elems.first()
        1

        Then get the next node.

        >>> elems.next()
        2

        """
        if self._itelm == -1:
            raise MapdlRuntimeError(
                "You first have to call the `DbElems.first` method."
            )

        request = mapdl_db_pb2.ElmRequest(next=self._itelm)
        result = self._db._stub.ElmNext(request)
        self._itelm = result.inum
        return self._itelm

    ###########################################################################
    # broken on server-side as of 2022R2 (server v0.5.1)
    # def next_defined(self):
    #     """get the number of the next defined elem
    #     You first have to call first)_

    #     Returns
    #     -------
    #     next_defined : int
    #     The next defined elem number
    #     = 0 - no more elems
    #     """

    #     if self._itelm == -1:
    #         raise TypeError("``elems.next`` you first have to call first function")

    #     request = mapdl_db_pb2.ElmRequest(next=self._itelm)
    #     result = self._db._stub.ElmNextDefined(request)
    #     self._itelm = result.inum
    #     return self._itelm
    ###########################################################################

    @check_mapdl_db_is_alive
    def info(self, ielm, ikey):
        """
        Get information about a element

        Parameters
        ----------
        ielm : int
            Element number. Should be 0 for key=11 for the following:

            * ``DB_NUMDEFINED``
            * ``DB_NUMSELECTED``
            * ``DB_MAXDEFINED``
            * ``DB_MAXRECLENG``

        ikey : int
            Key of the information needed about the node. One of the following:

            * DB_SELECTED : return select status

                * 0 - element is undefined.
                * -1 - element is unselected.
                *  1 - element is selected.

            * DB_NUMDEFINED - return number of defined elements
            * DB_NUMSELECTED - return number of selected elements
            * DB_MAXDEFINED - return highest element number defined
            * DB_MAXRECLENG - return maximum record length (dp words)
            * 2 - length (dp words)
            * 3 -
            * 4 - pointer to first data word
            * 11 - return void percent (integer)
            * 17 - pointer to start of index
            * 117 - return the maximum number of DP contact data stored for any element
            * -1 -
            * -2 - superelement flag
            * -3 - master dof bit pattern
            * -4 - active dof bit pattern
            * -5 - solid model attachment
            * -6 - pack nodal line parametric value
            * -7 - constraint bit pattern
            * -8 - force bit pattern
            * -9 - body force bit pattern
            * -10 - internal element flag
            * -11 - orientation element flag =1 is =0 isnot
            * -11 - contact element flag <0
            * -12 - constraint bit pattern (for DSYM)
            * -13 - if dof constraint written to file. (for LSDYNA only)
            * -14 - nodal coordinate system number (set by NROTATE)
            * -101 - pointer to element data record
            * -102 - pointer to angle record
            * -103 -
            * -104 - pointer to attached couplings
            * -105 - pointer to attacted constraint equations
            * -106 - pointer to nodal stresses
            * -107 - pointer to specified disp'S
            * -108 - pointer to specified forces
            * -109 - pointer to x/y/z record
            * -110 -
            * -111 -
            * -112 - pointer to nodal temperatures
            * -113 - pointer to nodal heat generations
            * -114 -
            * -115 - pointer to calculated displacements
            * -116 -

        Returns
        -------
        int
            The returned value of is based on the value of ``ikey``.

        Examples
        --------
        Query if an element is selected.

        >>> from ansys.mapdl.core.database import DBDef
        >>> elems = mapdl.db.elems
        >>> elems.info(1, DBDef.DB_SELECTED)
        1
        """
        if isinstance(ikey, DBDef):
            ikey = ikey.value

        request = mapdl_db_pb2.ElmIqrRequest(ielem=ielm, key=ikey)
        result = self._db._stub.ElmIqr(request)
        return result.elmiqr

    def num(self, selected=False):
        """
        Number of elements.

        Parameters
        ----------
        selected : bool, optional
            Return either the number of selected elements of the total number of
            elements.

        Returns
        -------
        int
            Number of elements. Set ``selected`` to ``True`` to include
            selection status.

        Examples
        --------
        Get the number of selected elements.

        >>> elems = mapdl.db.elems
        >>> elems.num(selected=True)
        425

        """
        if selected:
            return self.info(0, DBDef.DB_NUMSELECTED)
        return self.info(0, DBDef.DB_NUMDEFINED)

    @property
    def max_num(self) -> int:
        """
        Maximum element number.

        Examples
        --------
        Return the maximum element number.

        >>> elems = mapdl.db.elems
        >>> elems.max_num
        64

        """
        return self.info(0, DBDef.DB_MAXDEFINED.value)

    def get(self, ielm):
        """
        Get element attributes and nodes.

        Parameters
        ----------
        ielm : int
            Element number. Should be 0 for key=11 for the following:

        Returns
        -------
        ansys.api.mapdl.v0.mapdl_db_pb2.getelmResponse
            Element information containing the following attributes.

            * ``ielem`` : Element number
            * ``elmdat`` - Element data. Contains:

                * FIELD 0 : material reference number
                * FIELD 1 : element type number
                * FIELD 2 : real constant reference number
                * FIELD 3 : section number
                * FIELD 4 : element coordinate system
                * FIELD 5 : death flag (0 - alive, 1 - dead)
                * FIELD 6 : solid model reference
                * FIELD 7 : coded shape key
                * FIELD 8 : element number
                * FIELD 9 : base element number (applicable to reinforcing elements only)

            * ``nnod`` : Number of nodes
            * ``nodes`` : Node numbers belonging to the element

        Examples
        --------
        Return the element information of element 1.

        >>> elems = mapdl.db.elems
        >>> elem_info = elems.info(1)

        Return the nodes belonging to the element.

        >>> elem_info.nodes
        [2, 27, 37, 8]

        Return the element data.

        >>> elem_info.elmdat
        [1, 1, 1, 1, 0, 0, 14, 0, 0, 0]

        """
        request = mapdl_db_pb2.getelmRequest(ielem=ielm)
        return self._db._stub.getElm(request)

    def push(self, ielm, elmdat, nodes):
        """
        Push an element into the database.

        Examples
        --------
        Push a new linear hexahedron to MAPDL. This assumes that element type 1
        has been set to SOLID185 with ``mapdl.etype(1, "SOLID185")``.

        >>> elems = mapdl.db.elems
        >>> elems.push(
        ...     1,
        ...     [1, 1, 1, 1, 0, 0, 1, 0, 0, 0],
        ...     [1, 2, 3, 4, 5, 6, 7, 8],
        ... )

        """
        if len(elmdat) != 10:
            raise ValueError("`elmdat` must be length 10")
        request = mapdl_db_pb2.putelmRequest(
            ielem=ielm,  # int32
            elmdat=elmdat,  # repeated int32
            nnod=len(nodes),  # int32
            nodes=nodes,  # repeated int32
        )
        self._db._stub.putElm(request)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import IntEnum
from typing import Union
import warnings

from ansys.mapdl.core.errors import MapdlRuntimeError

QUERY_NAME = "__QUERY_PARM__"


class SelectionStatus(IntEnum):
    """Enumeration class for selection status information.

    This class is used with methods on the ``Query`` class and has the
    following options.

    UNSELECTED = -1
    UNDEFINED = 0
    SELECTED = 1

    Examples
    --------
    The following example is taken from ``Query.nsel`` and demonstrates
    how `SelectionSatus` appears in PyMAPDL.

    Here we create a single node and interrogate its selection
    status.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> from ansys.mapdl.core.inline_functions import Query
    >>> mapdl = launch_mapdl()
    >>> mapdl.prep7()
    >>> n1 = mapdl.n(1, 0, 0, 0)
    >>> n1
    1

    We can use ``Query.nsel`` to interrogate the selection status
    of the node. We can get the ``Query`` object from the
    ``mapdl.queries`` property. The response is an
    ``enum.IntEnum`` object. If you query a node that does not
    exist, it will return a status ``SelectionStatus.UNDEFINED``.

    >>> q = mapdl.queries
    >>> q.nsel(n1)
    <SelectionStatus.SELECTED: 1>
    >>> mapdl.nsel('NONE')
    >>> q.nsel(n1)
    <SelectionStatus.UNSELECTED: -1>
    >>> q.nsel(0)
    <SelectionStatus.UNDEFINED: 0>
    """

    UNSELECTED = -1
    UNDEFINED = 0
    SELECTED = 1


class _QueryExecution:
    def _run_query(self, command: str, integer: bool) -> Union[int, float]:
        # import here to avoid circular import
        from ansys.mapdl.core.mapdl_grpc import MapdlGrpc

        # non_interactive mode won't work with these commands
        if self._mapdl._store_commands:
            raise MapdlRuntimeError(
                "Inline MAPDL functions are incompatible with the "
                "non_interactive mode."
            )

        # use the underlying gRPC method if available to avoid parsing the string
        resp = self._mapdl._run(f"{QUERY_NAME}={command}")
        if isinstance(self._mapdl, MapdlGrpc):
            value = self._mapdl.scalar_param(QUERY_NAME)
            if value is None:
                raise MapdlRuntimeError(resp)
            if integer:
                return int(value)
            return value
        else:
            if integer:
                return self._parse_parameter_integer_response(resp)
            return self._parse_parameter_float_response(resp)

    def _parse_parameter_integer_response(self, response) -> int:
        """Parse integer response."""
        return int(self._parse_parameter_float_response(response))

    @staticmethod
    def _parse_parameter_float_response(response) -> float:
        if "PARAMETER" not in response or "=" not in response:
            raise TypeError(f"Parameter response not recognised: " f'"{response}"')
        parts = response.rsplit("=", 1)
        if "WARNING" in parts[0]:
            warnings.warn(parts[0])

        number = parts[1].strip()
        return float(number)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .core import _QueryExecution


class _LineFractionCoordinateQueries(_QueryExecution):
    _mapdl = None

    def lx(self, n: int, lfrac: float) -> float:
        """X-coordinate of line ``n`` at length fraction ``lfrac``.

        Fetches X-coordinate of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.

        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The X-coordinate.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 3)`` then find the X-coordinate halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 3)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.lx(l0, 0.5)
        0.5
        """
        return self._run_query(f"LX({n}, {lfrac})", integer=False)

    def ly(self, n: int, lfrac: float) -> float:
        """Y-coordinate of line ``n`` at length fraction ``lfrac``.

        Fetches Y-coordinate of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.

        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The Y-coordinate.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 3)`` then find the Y-coordinate halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 3)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.ly(l0, 0.5)
        1.0
        """
        return self._run_query(f"LY({n}, {lfrac})", integer=False)

    def lz(self, n: int, lfrac: float) -> float:
        """Z-coordinate of line ``n`` at length fraction ``lfrac``.

        Fetches Z-coordinate of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.

        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The Z-coordinate.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 3)`` then find the Z-coordinate halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 3)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.lz(l0, 0.5)
        1.5
        """
        return self._run_query(f"LZ({n}, {lfrac})", integer=False)


class _LineFractionSlopeQueries(_QueryExecution):
    _mapdl = None

    def lsx(self, n: int, lfrac: float) -> float:
        """X-slope of line ``n`` at length fraction ``lfrac``.

        Fetches X-slope of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.
        This is equivalent to the rate of change of the x-coordinate
        with respect to the change of line length.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.
        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The X-slope.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 2)`` then find the X-slope halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 2)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.lsx(l0, 0.5)
        0.3333333333
        """
        return self._run_query(f"LSX({n}, {lfrac})", integer=False)

    def lsy(self, n: int, lfrac: float) -> float:
        """Y-slope of line ``n`` at length fraction ``lfrac``.

        Fetches Y-slope of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.
        This is equivalent to the rate of change of the y-coordinate
        with respect to the change of line length.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.
        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The Y-slope.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 2)`` then find the Y-slope halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 2)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.lsy(l0, 0.5)
        0.6666666667
        """
        return self._run_query(f"LSY({n}, {lfrac})", integer=False)

    def lsz(self, n: int, lfrac: float) -> float:
        """Z-slope of line ``n`` at length fraction ``lfrac``.

        Fetches Z-slope of line ``n`` at ``lfrac`` times the line
        length along the line. ``0. <= lfrac <= 1.``.
        This is equivalent to the rate of change of the z-coordinate
        with respect to the change of line length.

        Parameters
        ----------
        n : int
            The line number of the line to be considered.
        lfrac: float
            The fraction of the length of the line along which to
            enquire. ``0. <= lfrac <= 1.``

        Returns
        -------
        float
            The Z-slope.

        Examples
        --------
        Here we construct a line between the coordinates ``(0, 0, 0)``
        and ``(1, 2, 2)`` then find the Z-slope halfway along the line.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.prep7()
        >>> k0 = mapdl.k(1, 0, 0, 0)
        >>> k1 = mapdl.k(2, 1, 2, 2)
        >>> l0 = mapdl.l(k0, k1)
        >>> q = mapdl.queries
        >>> q.lsz(l0, 0.5)
        0.6666666667
        """
        return self._run_query(f"LSZ({n}, {lfrac})", integer=False)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""These commands may be run by PyMAPDL, but are not supported from a
user-context.

"""


class _Hidden:
    def _batch(self, lab="", **kwargs):
        """APDL Command: /BATCH

        Sets the program mode to "batch."

        Parameters
        ----------
        lab
            Specifies listing mode during a batch run:

            LIST
                The batch output will include a listing of the input file.

            (blank)
                Suppress input data listing.

        Notes
        -----
        Sets the program mode to "batch" when included as the first line on an
        input file of ANSYS commands.  For convenience, this command is
        automatically recorded on the log file (Jobname.LOG) at the beginning
        of an interactive ANSYS session so that the log file can be reused
        later for batch input.

        Caution:: : This command  should not be entered directly in an
        interactive ANSYS session since all subsequent keyboard input is simply
        copied to a file, without further processing or prompts from the
        program (use the "system break" to exit the ANSYS program if this
        occurs).

        The recommended method for choosing batch mode, rather than using the
        /BATCH command, is to select the Batch simulation environment from the
        ANSYS Product Launcher  task in the ANSYS launcher, or the batch mode
        entry option on the ANSYS execution command when entering the program.

        This command is valid only at the Begin Level.
        """
        command = "/BATCH,%s" % (str(lab))
        return self.run(command, **kwargs)

    def _output(self, fname="", ext="", loc="", **kwargs):
        """Redirects text output to a file or to the screen.

        APDL Command: /OUTPUT

        Parameters
        ----------
        fname
            Filename and directory path (248 character maximum, including
            directory) to which text output will be redirected (defaults to
            Jobname if Ext is specified).  For interactive runs, Fname = TERM
            (or blank) redirects output to the screen.  For batch runs, Fname =
            blank (with all remaining command arguments blank) redirects output
            to the  default system output file.

        ext
            Filename extension (eight-character maximum).

        loc
            Location within a file to which output will be written:

            (blank)
                Output is written starting at the top of the file (default).

            APPEND
                Output is appended to the existing file.

        Notes
        -----
        Text output includes responses to every command and GUI
        function, notes, warnings, errors, and other informational
        messages.  Upon execution of /OUTPUT,Fname, Ext, ..., all
        subsequent text output is redirected to the file Fname.Ext.
        To redirect output back to the default location, issue /OUTPUT
        (no arguments).

        This command is valid in any processor.
        """
        if loc:
            return self.run(f"/OUTPUT,{fname},{ext},,{loc}", **kwargs)
        else:
            return self.run(f"/OUTPUT,{fname},{ext}", **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class MultiFieldSolverTimeControls:
    def mfcalc(self, fnumb="", freq="", **kwargs):
        """Specifies a calculation frequency for a field in an ANSYS Multi-field

        APDL Command: MFCALC
        solver analysis.

        Parameters
        ----------
        fnumb
            Field number set by the MFELEM command.

        freq
            Perform calculation every Nth ANSYS Multi-field solver time step.
            Defaults to 1 for every time step.

        Notes
        -----
        This command only applies to a harmonic analysis of the specified
        field. It is useful when a field contributes negligible field
        interaction within a single ANSYS Multi-field solver time step.

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFCALC,{fnumb},{freq}"
        return self.run(command, **kwargs)

    def mfdtime(self, dtime="", dtmin="", dtmax="", carry="", **kwargs):
        """Sets time step sizes for an ANSYS Multi-field solver analysis.

        APDL Command: MFDTIME

        Parameters
        ----------
        dtime
            Multi-field time step size.  If automatic time stepping is being
            used [see Notes below], DTIME is the starting time step.

        dtmin
            Minimum time step. Defaults to DTIME.

        dtmax
            Maximum time step. Defaults to DTIME.

        carry
            Time step carryover key.

            OFF  - Use DTIME as the starting time step for the next restart run (default).

            ON  - Use the final time step from the previous run as the starting time step for the
                  next restart run.

        Notes
        -----
        This command specifies time step sizes for an ANSYS Multi-field solver
        analysis. If either DTMIN or DTMAX is not equal to DTIME, auto time-
        stepping is turned on for the multi-field loop. ANSYS will
        automatically adjust the time step size for the next multi-field step
        between DTMIN and DTMAX, based on the status of the current
        convergence, the number of target stagger iterations (specified by
        MFITER), and the actual number of iterations needed to reach
        convergence at the current time step.

        If auto time-stepping is turned off, the time step size must be evenly
        divisible into the end time (specified by MFTIME) minus the start time
        (0 for a new analysis or a restart time specified by MFRSTART).

        You can use a smaller time step within each ANSYS field analysis. This
        is called subcycling. Use the DELTIM and AUTOTS commands to subcycle a
        structural, thermal, or electromagnetic analysis.

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFDTIME,{dtime},{dtmin},{dtmax},{carry}"
        return self.run(command, **kwargs)

    def mfoutput(self, freq="", **kwargs):
        """Specifies results file output frequency for an ANSYS

        APDL Command: MFOUTPUT
        Multi-field solver analysis.

        Parameters
        ----------
        freq
            N

            N - Write solution every Nth (and the last) time
                 step. Defaults to 1, for every time step.

            -N - Writes up to -N equally spaced results (for multifield auto time stepping).

            NONE - Suppresses writing of results for all multifield time steps.

            ALL - Writes results for every multifield time step (default).

            LAST - Writes results for only the last multifield time step.

            %array% - Where %array% is the name of an n X 1 X 1
                      dimensional array parameter defining n key
                      times, the data for the specified solution
                      results item is written at those key times. Key
                      times in the array parameter must appear in
                      ascending order. Value must be greater than or
                      equal to the ending time values for the load
                      step.

            For restart runs (see MFRSTART command), either change the
            parameter values to fall between the beginning and ending
            time values of the load step, or erase the current
            settings and reissue the command with a new array
            parameter.  - For more information about defining array
            parameters, see the ``*DIM`` command documentation.

        Notes
        -----
        A MFOUTPUT setting overrides any other output frequency setting
        (OUTRES). To select the solution items, use the OUTRES command.

        For the case of Freq = -n and Freq = %array%, the results at the time
        points which first time reaches or exceeds the targeting ouptupt time
        points will be written.

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFOUTPUT,{freq}"
        return self.run(command, **kwargs)

    def mfrstart(self, time="", **kwargs):
        """Specifies restart status for an ANSYS Multi-field solver analysis.

        APDL Command: MFRSTART

        Parameters
        ----------
        time
            Restart time

            0  -  New analysis (Default)

            -1  - Restart from the last result set from a previous run.

            n  - Specify any positive number for the actual time point from which the ANSYS
                 Multi-field solver will restart. ANSYS checks the availability
                 of the result set and database file.

        Notes
        -----
        For MFX analyses, ANSYS always passes an actual time value to CFX (zero
        for a new analysis or a positive value for a restart run) and CFX
        verifies the consistency with the initial results file. For more
        details about ANSYS restart capabilities, please see Restarting an
        Analysis in the Basic Analysis Guide.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFRSTART,{time}"
        return self.run(command, **kwargs)

    def mftime(self, time="", **kwargs):
        """Sets end time for an ANSYS Multi-field solver analysis.

        APDL Command: MFTIME

        Parameters
        ----------
        time
            End time of an ANSYS Multi-field solver analysis. Defaults to 1.

        Notes
        -----
        A MFTIME setting overrides any other end time setting (TIME).

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFTIME,{time}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import weakref

import numpy as np

from ansys.mapdl.core import Mapdl
from ansys.mapdl.core.errors import MapdlRuntimeError


class KrylovSolver:
    """Abstract mapdl krylov class.  Created from a ``Mapdl`` instance.

    Notes
    -----
    The procedure to use the Krylov solver is composed of three steps:

    1. Generate Krylov subspace using
       :func:`KrylovSolver.gensubspace <ansys.mapdl.core.krylov.KrylovSolver.gensubspace>`.

    2. Use :func:`KrylovSolver.solve <ansys.mapdl.core.krylov.KrylovSolver.solve>`
       to solve the generated Krylovsub space using a reduced harmonic analysis over
       a specified frequency range

    3. Expand the reduced solution back to the original space using
       :func:`KrylovSolver.solve <ansys.mapdl.core.krylov.KrylovSolver.solve>`

    Examples
    --------
    Create an instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> ....
    >>> ....
    >>> Generate the FEA model (mesh, constraints, loads)
    >>> Generate the .full file

    >>> mk = mapdl.krylov
    >>> # Generate the Krylov subspace
    >>> Qz = mk.gensubspace(10, 100, check_orthogonality=True)
    >>> # Reduce the system of equations and solve at each frequency.
    >>> Yz = mk.solve(10, 100,  freq_steps=1, ramped_load=True)
    >>> # Expand the reduced solution back to the FE space.
    >>> res = mk.expand(residual_computation=True, residual_algorithm="l2")

    """

    def __init__(self, mapdl: Mapdl):
        """Krylov analysis

        Abstract mapdl krylov class.  Created from an
        :class:`Mapdl instance <ansys.mapdl.core.Mapdl>` instance.

        Parameters
        ----------
        mapdl : ansys.mapdl.core.Mapdl
            Mapdl instance which this class references to.
        """
        if not isinstance(mapdl, Mapdl):  # pragma: no cover
            raise TypeError("``mapdl`` must be an Mapdl class instance")
        self._mapdl_weakref = weakref.ref(mapdl)
        self.mm = self._mapdl.math
        self.jobname = self._mapdl.jobname
        self.logger = self._mapdl._log

        # run flags
        self._run_gensubspace = False
        self._run_solve = False

        self.residuals = None
        self.solution_vectors = None
        self.orthogonality = None

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl."""
        return self._mapdl_weakref()

    def _check_full_file_exists(self, full_file=None):
        """Check full file exists."""
        current_dir = self._mapdl.directory
        # Check if full file exists
        if not full_file:
            self.full_file = self._mapdl.jobname + ".full"
        else:
            self.full_file = full_file

        # Checking if the full file exists.
        if self._mapdl._local:
            if not os.path.exists(os.path.join(current_dir, self.full_file)):
                raise FileNotFoundError(
                    f"The file '{self.full_file}' could not be found in local directory '{current_dir}'."
                )
        else:
            if self.full_file not in self._mapdl.list_files():
                raise FileNotFoundError(
                    f"The file '{self.full_file}' could not be found in the remote MAPDL instance."
                )

    @property
    def is_orthogonal(self):
        """
        Check whether the solution is orthogonal.

        Returns
        -------
        bool
            If the matrix is orthogonal, returns ``True``, otherwise is ``False``.
        """
        if self.orthogonality is not None:
            eye_ = np.eye(N=self.orthogonality.shape[0])
            return np.allclose(self.orthogonality, eye_)

    def _check_input_gensubspace(self, max_dim_q, freq_val, check_orthogonality):
        """Validate the inputs to the ``gensubspace`` method."""

        # Check for illegal input values by the user
        if not isinstance(max_dim_q, int) or max_dim_q <= 0:
            raise ValueError(
                "The maximum size of the Krylov subspace must be greater than 0."
            )

        if not isinstance(freq_val, int) or freq_val < 0:
            raise ValueError(
                "The frequency value ('freq_val') for building the Krylov subspace must be "
                "equal to or greater than 0 Hz."
            )

        if not isinstance(check_orthogonality, bool):
            raise ValueError(
                "The 'check_orthogonality' value for building the Krylov subspace must be "
                "True or False"
            )

    def _check_input_solve(self, freq_start, freq_end, freq_steps, ramped_load):
        """Validate the inputs to the ``solve`` method."""

        if not isinstance(freq_start, int) or freq_start < 0:
            raise ValueError(
                "The beginning frequency value for solving the reduced solution must be "
                "equal to or greater than 0 Hz."
            )

        if not isinstance(freq_end, int) or freq_end < 0 or freq_end < freq_start:
            raise ValueError(
                "The beginning frequency value for solving the reduced solution must be "
                "equal to or greater than 0 Hz, and greater than starting frequency.'"
            )

        if not isinstance(freq_steps, int) or freq_steps < 1:
            raise ValueError(
                "The number of frequencies ('freq_steps') for which to compute the reduced "
                "solution must be to be an integer greater than or equal to 1."
            )

        if not isinstance(ramped_load, bool):
            raise ValueError(
                "The 'ramped_load' argument for computing the reduced solution must be "
                "True or False"
            )

    def _check_input_expand(
        self, return_solution, residual_computation, residual_algorithm
    ):
        """Validate the inputs to the ``expand`` method."""

        if not isinstance(return_solution, bool):
            raise ValueError(
                "The 'return_solution' value for expanding the solution must be True or False"
            )
        if not isinstance(residual_computation, bool):
            raise ValueError("The 'residual_computation' must be True or False.")
        if not isinstance(
            residual_algorithm, str
        ) or residual_algorithm.lower() not in [
            "l-inf",
            "linf",
            "l-1",
            "l1",
            "l-2",
            "l2",
        ]:
            raise ValueError(
                "The provided 'residual_algorithm' is not allowed. Only allowed are 'L-inf', 'Linf', 'L-1', 'L1', 'L-2', 'L2'."
            )

    def _get_data_from_full_file(self):
        """Extract stiffness, mass, damping, and force matrices from the FULL file."""

        self._mat_k = self.mm.stiff(fname=self.full_file)
        self._ndof = self._mat_k.shape[1]
        self._mat_m = self.mm.mass(fname=self.full_file)
        self._mat_c = self.mm.damp(fname=self.full_file)

        self._mapdl.smat("Nod2Solv", "D", "IMPORT", "FULL", self.full_file, "NOD2SOLV")
        self.Nod2Solv = self.mm.mat(name="Nod2Solv")

        self._mapdl.vec("fz", "Z", "IMPORT", "FULL", self.full_file, "RHS")
        self._mapdl.vec("fz0", "Z", "COPY", "fz")
        self.fz0 = self.mm.vec(name="fz0")

    def _calculate_orthogonality(self, uz, num_q):
        """Check Orthonormality of vectors"""

        if self.orthogonality is not None:
            # Using previously calculated (if any)
            return self.orthogonality

        orthogonality = np.zeros(shape=(num_q, num_q))
        for i in range(num_q):
            self._mapdl.vec("Vz", "Z", "LINK", self.Qz.id, i + 1)
            Vz = self.mm.vec(name="Vz")
            Vz_m = Vz.asarray()
            for j in range(num_q):
                self._mapdl.vec(uz.id, "Z", "LINK", self.Qz.id, j + 1)
                uz_m = uz.asarray()
                rcon = np.vdot(Vz_m, uz_m)
                rcon_real = rcon.real
                rcon_imag = rcon.imag
                dcon = np.sqrt(rcon_real * rcon_real + rcon_imag * rcon_imag)
                orthogonality[i, j] = dcon

        self.orthogonality = orthogonality
        return self.orthogonality

    def gensubspace(
        self, max_dim_q, frequency, check_orthogonality=False, full_file=None
    ):
        """Generate a Krylov subspace for model reduction in a harmonic analysis.

        This method generates a Krylov subspace used for a model reduction
        solution in a harmonic analysis. The subspace is built using the
        assembled matrices and load vector on the ``<jobname>.full`` file that
        is located in the current working directory. This FULL file
        should be built at the specified frequency value.

        Parameters
        ----------
        max_dim_q : int
          Maximum size/dimension of the Krylov subspace.

        frequency : int
          Frequency value in Hz to build the Krylov subspace at.

        check_orthogonality : bool, optional
          Whether to check the orthonormal properties of each subspace vector
          with all other subspace vectors. The result matrix is stored in
          :attr:`KrylovSolver.orthogonality <ansys.mapdl.core.krylov.KrylovSolver.orthogonality`.
          The default is ``False``.

        full_file : str, optional
          Name of the FULL file to read. The default is ``<jobname>.full``.

        Returns
        -------
        AnsMath
            Krylov subspace.

        Notes
        -----
        Distributed Ansys restriction: This command is not supported in
        Distributed Ansys.
        """

        current_dir = self._mapdl.directory
        self.logger.debug(f"Current directory: {current_dir}")

        two_pi = 2.0 * np.pi
        omega_zero = frequency * two_pi
        ccf = -(omega_zero**2)
        init_val = 2  # Initialized as no damping
        zero_c = 0

        # Check full file exists
        self._check_full_file_exists(full_file)

        # Check the inputs
        self._check_input_gensubspace(max_dim_q, frequency, check_orthogonality)

        # Get matrices from full file
        self._get_data_from_full_file()

        # Get the force vector from the defined AnsVec
        fz = self.mm.vec(name="fz")

        # Create subspace
        self.Qz = self.mm.zeros(self._ndof, max_dim_q, dtype=np.cdouble)

        # Form az = (K-w0*w0*M,i*w0*C)
        self._mapdl.smat("az", "Z", "COPY", self._mat_k.id)
        az = self.mm.mat(name="az")
        az.axpy(self._mat_m, ccf, 1.0)
        self._mapdl.axpy(0.0, omega_zero, self._mat_c.id, 1.0, 0.0, az.id)

        # Form cz=(C+i*2*w0*M)
        self._mapdl.smat("cz", "Z", "COPY", self._mat_c.id)
        cz = self.mm.mat(name="cz")
        self._mapdl.axpy(0.0, 2 * omega_zero, self._mat_m.id, 1.0, 0.0, cz.id)

        norm_c = self._mapdl.nrm(cz.id, "NRM2", "normC", "NO")

        if not norm_c:
            zero_c = 1

        # Solve for the 1st vector of subspace[Q]
        # Create solver system of [az]
        s = self.mm.factorize(az)  # Factor [az]

        # {uz1} linked to Qz[1]
        self._mapdl.vec("uz1", "Z", "LINK", self.Qz.id, 1)
        uz1 = self.mm.vec(name="uz1")

        uz1 = s.solve(fz, uz1)
        norm_u = self._mapdl.nrm(uz1.id, "NRM2", "normU", "YES")

        if not zero_c:
            init_val = 3
            self._mapdl.vec("uz2", "Z", "LINK", self.Qz.id, 2)
            uz2 = self.mm.vec(name="uz2")
            fz.zeros()  # {fz}=0

            # {fz}={fz}+[cz]{uz1}
            fz = cz.dot(uz1)
            fz *= -1  # {fz}=-{fz}
            norm_u = self._mapdl.nrm(fz.id, "NRM2", "normU", "YES")

            uz2 = s.solve(fz, uz2)  # back solved:{uz2}=[az]^-1*{fz}

            # {uz2}*{v1} = ccon
            self._mapdl.dot(uz1.id, uz2.id, "ccon_real", "ccon_imag")
            # {uz2}={uz2}-ccon*{uz1}
            self._mapdl.axpy("-ccon_real", "-ccon_imag", uz1.id, 1.0, 0.0, uz2.id)
            # {uz2} normalized
            norm_u = self._mapdl.nrm(uz2.id, "NRM2", "normU", "YES")

        # Build Subspace Vectors with MGS (2nd Order)
        for dim_q in range(init_val, max_dim_q + 1):
            self.dim_q = dim_q
            if norm_u == 0:
                num_q = dim_q - 1  # exhausted at dim_q
            else:
                num_q = max_dim_q  # may reach to the maximum

            if num_q >= max_dim_q:  # make next vector
                fz.zeros()
                if zero_c == 0:  # [cz]!=0 case
                    next_uz_one = dim_q - 2
                    next_uz_two = dim_q - 1
                    # {uz1} to [Qz](dim_q-2)
                    self._mapdl.vec(uz1.id, "Z", "LINK", self.Qz.id, next_uz_one)
                    # {uz2} to [Qz](dim_q-1)
                    self._mapdl.vec(uz2.id, "Z", "LINK", self.Qz.id, next_uz_two)
                    # {uz} to [Qz](dim_q)
                    self._mapdl.vec("uz", "Z", "LINK", self.Qz.id, dim_q)
                    uz = self.mm.vec(name="uz")
                    # [cz]{uz2} -> {uz}
                    uz += cz.dot(uz2)

                else:  # [cz]=0 case
                    next_uz_one = dim_q - 1
                    self._mapdl.vec(uz1.id, "Z", "LINK", self.Qz.id, next_uz_one)
                    # {uz} to [Qz](dim_q)
                    self._mapdl.vec(uz2.id, "Z", "LINK", self.Qz.id, dim_q)
                    uz.zeros()

                # [MatM]{uz1} -> {fz}
                self._mapdl.mult(m1=self._mat_m.id, t1="", m2=uz1.id, t2="", m3=fz.id)
                fz.axpy(uz, -1.0, -1.0)  # {fz} = -{fz}-{uz}
                uz = s.solve(fz, uz)  # [az]^-1{fz} -> {uz}

                # Make subspace vectors orthonormal
                for j in range(1, dim_q):
                    self._mapdl.vec("V1", "Z", "LINK", self.Qz.id, j)
                    V1 = self.mm.vec(name="V1")
                    self._mapdl.dot(V1.id, uz.id, "ccon_real", "ccon_imag")
                    self._mapdl.axpy("-ccon_real", "-ccon_imag", V1.id, 1.0, 0.0, uz.id)

                norm_u = self._mapdl.nrm(uz.id, "NRM2", "normU", "YES")

        # Optional check on Orthonormality of vectors
        if check_orthogonality:
            self._calculate_orthogonality(uz, num_q)

        self._run_gensubspace = True
        return self.Qz

    def solve(self, freq_start, freq_end, freq_steps, ramped_load=True):
        """Reduce the system of equations and solve at each frequency.

        This method uses a Krylov subspace to solve a reduced harmonic
        analysis over a specified frequency range for a given number of
        frequency points (intervals).

        Parameters
        ----------
        freq_start : int
          Starting value of the frequency range in Hz.
        freq_end : int
          Ending value of the frequency range in Hz.
        freq_steps : int
          Number of intervals in the frequency range.
        ramped_load : bool
          Whether to use ramped load. If not, the load used is stepped (``False``).
          Defaults to ``True``.

        Returns
        -------
        AnsMath
            Reduced solution over the frequency range.

        Notes
        -----
        Distributed Ansys restriction: This command is not supported in
        Distributed Ansys.
        """

        # Check we ran gensubspace method before
        if not self._run_gensubspace:
            raise MapdlRuntimeError(
                "The method 'gensubspace' should be executed first."
            )

        # Check inputs before executing the method
        self._check_input_solve(freq_start, freq_end, freq_steps, ramped_load)

        # Execute macro if no errors raised
        # Store input arguments from user
        self.freq_start = freq_start
        self.freq_steps = freq_steps
        self.ramped_load = ramped_load
        self.freq_end = freq_end

        az = self.mm.mat(name="az")
        ndof = self._ndof
        dim_q = self.dim_q

        # Define solution variables
        QtF = self.mm.zeros(dim_q, dtype=np.cdouble)
        QtAQ = self.mm.zeros(dim_q, dim_q, dtype=np.cdouble)  # [QtAQ], reduced matrix
        self.Yz = self.mm.zeros(
            dim_q, self.freq_steps, dtype=np.cdouble
        )  # [Yz], reduced solution over range
        self.DzV = self.mm.zeros(
            ndof, self.freq_steps, dtype=np.cdouble
        )  # {DzV}, displacement vector
        self.iRHS = self.mm.zeros(ndof, dtype=np.cdouble)

        # Loop over frequency range
        omega = self.freq_start * 2 * np.pi
        self.intV = (self.freq_end - self.freq_start) / self.freq_steps

        for iFreq in range(1, self.freq_steps + 1):
            # form RHS at the i-th frequency point
            self.iRHS.zeros()
            if self.ramped_load == 0:
                # apply ramped loading
                ratio = 2 * iFreq / self.freq_steps
                if self.freq_steps == 1:
                    ratio = 1.00
            else:
                # Apply stepped loading
                ratio = 1.00

            self.iRHS.axpy(self.fz0, ratio, 1.0)  # Get {iRHS} at iFreq
            QtF = self.Qz.T.dot(self.iRHS)  # Reduce [Qz]^t{iRHS} -> {QtF}

            # link iY to Yz(iFreq)
            self._mapdl.vec("iY", "Z", "LINK", self.Yz.id, iFreq)
            self.iY = self.mm.vec(name="iY")

            # Reduced system at omega
            omega = omega + self.intV * 2 * np.pi  # defined frequency value
            ccf = -omega * omega

            # modified K:[az]=[MatK]
            self._mapdl.smat("az", "Z", "COPY", self._mat_k.id)
            az = self.mm.mat(name="az")

            # [az]=(1,0)*[az]+(ccf,0)*[MatM]
            az.axpy(self._mat_m, ccf, 1.0)
            # [az]=(1,0)*[az]-(0,omega)*[mat_C]
            self._mapdl.axpy(0.0, omega, self._mat_c.id, 1.0, 0.0, az.id)

            AQ = az.dot(self.Qz)  # [AQ]=[az][Qz]
            QtAQ = self.Qz.T.dot(AQ)  # [QtAQ]=[Q^t][AQ]

            # create reduced system of equations
            s = self.mm.factorize(QtAQ)
            # back solve for {iY}=Yz[iFreq]
            self.iY = s.solve(QtF, self.iY)

        self._run_solve = True
        return self.Yz

    def expand(
        self,
        residual_computation=False,
        residual_algorithm=None,
        compute_solution_vectors=True,
        return_solution=False,
    ):
        """Expand the reduced solution back to FE space.

        This method expands the reduced solution for a harmonic analysis
        back to the original space. Optional calculation of the residual
        is available.

        Parameters
        ----------
        residual_computation  : bool, optional
          Compute the residual of the expanded solution.
          Default to ``False``.

        residual_algorithm : str, optional
          Specifies the type of residual normal calculation. It can take
          the following values:
          * "L-inf": Compute the L-inf norm of the residual. Default value.
          * "L-1": Compute the L-1 norm of the residual.
          * "L-2": Compute the L-2 norm of the residual.

        compute_solution_vectors : bool, optional
          If ``True`` it compute the solution vectors. The solution vectors
          are stored in :attr:`Krylov.solution_vectors
          <ansys.mapdl.core.krylov.KrylovSolver.solution_vectors>` method.
          Defaults to ``True``.

        return_solution : bool, optional
          If ``True`` it will return the solution vectors. By default is ``False``.

        Returns
        -------
        np.ndarray
            Solution vectors mapped to a given order. (if ``return_solution = True``)
        None
            If ``return_solution = False``

        Notes
        -----
        Distributed Ansys restriction: This command is not supported in
        Distributed Ansys.
        """
        # Check we ran solve method before
        if not self._run_solve:
            raise MapdlRuntimeError("The method 'solve' should be executed first.")

        if residual_algorithm is not None:
            # To avoid having to set residual computation.
            residual_computation = True
        elif residual_algorithm is None:
            residual_algorithm = "L-Inf"

        # Check inputs before executing the method
        self._check_input_expand(
            return_solution, residual_computation, residual_algorithm
        )

        self._residual_algorithm = residual_algorithm

        RzV = self.mm.zeros(self._ndof, 1, dtype=np.cdouble)

        # Build mapping vectors
        self._mapdl.vec("MapForward", "I", "IMPORT", "FULL", self.full_file, "FORWARD")
        self._map_forward = self.mm.vec(name="MapForward")

        self._mapdl.vec("MapBack", "I", "IMPORT", "FULL", self.full_file, "BACK")
        map_back = self.mm.vec(name="MapBack")

        self._max_node = self._map_forward.size
        self._num_node = map_back.size

        # Expand reduced solution
        omega = self.freq_start * 2 * np.pi  # circular frequency at starting point
        xii_usr_ordered = []

        for iFreq in range(1, self.freq_steps + 1):
            self._mapdl.vec(self.iY.id, "Z", "LINK", self.Yz.id, iFreq)
            # vector of solution in solver space
            self._mapdl.vec("Xi", "Z", "LINK", self.DzV.id, iFreq)
            Xi = self.mm.vec(name="Xi")

            Xi.zeros()
            # collect each vector's contribution
            for jVect in range(1, self.dim_q + 1):
                self._mapdl.vec("Qzj", "Z", "LINK", self.Qz.id, jVect)
                Qzj = self.mm.vec(name="Qzj")

                iY_a = self.iY.asarray()
                yrj = iY_a[jVect - 1].real  # real part of reduced solution
                yij = iY_a[jVect - 1].imag  # imaginary part of reduced solution
                self._mapdl.axpy(yrj, yij, Qzj.id, 1.0, 0.0, Xi.id)

            # Output solution vectors (if requested)
            if compute_solution_vectors:
                dof_each_freq = self._compute_solution_vector(Xi)
                xii_usr_ordered.append(dof_each_freq)

            # Compute residual norm (if requested)
            if residual_computation or residual_algorithm.lower() in [
                "no",
                "off",
            ]:
                norm_rz, norm_fz = self.compute_residuals(iFreq, RzV, Xi, omega)
                if not self.residuals:
                    self.residuals = []

                self.residuals.append([iFreq, norm_rz, norm_fz])

        # Storing solution in class
        if compute_solution_vectors:
            self.solution_vectors = np.array(
                xii_usr_ordered,
                dtype=[("node", "i4"), ("equ", "i4"), ("x", "c8")],
            )

        if return_solution:
            return self.solution_vectors

    def compute_residuals(self, iFreq, RzV, Xi, omega):
        """Compute residuals of the matrices"""
        # form {iRHS}
        self.iRHS.zeros()
        if self.ramped_load:
            # apply ramped loading
            ratio = 2 * iFreq / self.freq_steps
            if self.freq_steps == 1:
                ratio = 1.0
        else:
            # Apply stepped loading
            ratio = 1.0

        self.iRHS.axpy(self.fz0, ratio, 1.0)  # get {iRHS} at iFreq

        # Form az
        omega = omega + self.intV * 2 * np.pi  # defined frequency value
        ccf = -omega * omega

        self._mapdl.smat("az", "Z", "COPY", self._mat_k.id)  # modified K:[az]=[MatK]
        az = self.mm.mat(name="az")

        # [az]=(1,0)*[az]+(ccf,0)*[MatM]
        az.axpy(self._mat_m, ccf, 1.0)
        # [az]=(1,0)*[az]-(0,omega)*[mat_C]
        self._mapdl.axpy(0.0, omega, self._mat_c.id, 1.0, 0.0, az.id)

        # Compute {Rz}={iRHS}-[az]*{Xi}
        self._mapdl.vec("Rzi", "Z", "LINK", RzV.id, 1)
        Rzi = self.mm.vec(name="Rzi")

        Rzi.zeros()
        Rzi = az.dot(Xi)
        Rzi.axpy(self.iRHS, 1.0, -1.0)

        # Output norms of residual vector
        norm_rz = 0.0

        if self._residual_algorithm.lower() in ["l-inf", "linf"]:  # L-inf norm
            norm_rz = Rzi.norm("NRMINF")
            norm_fz = self.iRHS.norm("NRMINF")
            if norm_fz != 0.0:
                norm_rz = norm_rz / norm_fz

        elif self._residual_algorithm.lower() in ["l-1", "l1"]:
            norm_rz = Rzi.norm("NRM1")
            norm_fz = self.iRHS.norm("NRM1")
            if norm_fz != 0.0:
                norm_rz = norm_rz / norm_fz

        elif self._residual_algorithm.lower() in ["l-2", "l2"]:
            norm_rz = Rzi.norm("NRM2")
            norm_fz = self.iRHS.norm("NRM2")
            if norm_fz != 0.0:
                norm_rz = norm_rz / norm_fz
        else:
            raise ValueError(
                f"The provided algorithm '{self._residual_algorithm}' is not supported."
            )

        return norm_rz, norm_fz

    def _compute_solution_vector(self, Xi):
        self._mapdl.mult(
            m1=self.Nod2Solv.id, t1="TRANS", m2=Xi.id, t2="", m3="Xii"
        )  # Map {Xi} to internal (ANSYS) order
        Xii = self.mm.vec(name="Xii")
        dof_each_freq = []

        # Map {Xii} to user ordering
        num_eqn = Xii.size  # dim 1
        numdof = int(num_eqn / self._num_node)

        for extnode in range(1, self._max_node + 1):
            intNode = self._map_forward[extnode - 1]
            if intNode < 0:
                break
            intEqn = int((intNode - 1) * numdof)
            extEqn = int((extnode - 1) * numdof)

            for idof in range(1, numdof + 1):
                Xii_real = Xii.asarray()[intEqn + idof - 1].real  # Real value
                Xii_imag = Xii.asarray()[intEqn + idof - 1].imag  # Imag value

                dof_each_freq.append(
                    (int(extnode), int(extEqn + idof), Xii_real + Xii_imag * 1j)
                )
        return dof_each_freq
